#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import getpass
import logging
import os
import shutil
import tempfile
import traceback
import zipfile
import pkg_resources
from concurrent.futures.thread import ThreadPoolExecutor
from io import BytesIO, StringIO

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s')

import dacite
import pandas as pd
from flask import Flask, jsonify, request, send_file, make_response, send_from_directory, current_app, Blueprint
from flask_cors import CORS, cross_origin

from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager
from label_sleuth.training_set_selector.training_set_selector_factory import TrainingSetSelectionFactory
from label_sleuth.app_utils import elements_back_to_front, extract_iteration_information_list, \
    extract_enriched_ngrams_and_weights_list, get_element, get_natural_sort_key, validate_category_id, \
    validate_workspace_id
from label_sleuth.authentication import authenticate_response, login_if_required, verify_password
from label_sleuth.active_learning.core.active_learning_factory import ActiveLearningFactory
from label_sleuth.config import Configuration
from label_sleuth.configurations.users import User
from label_sleuth.data_access.core.data_structs import LABEL_POSITIVE, LABEL_NEGATIVE, Label
from label_sleuth.data_access.data_access_api import AlreadyExistsException
from label_sleuth.data_access.file_based.file_based_data_access import FileBasedDataAccess
from label_sleuth.models.core.models_factory import ModelFactory
from label_sleuth.models.core.tools import SentenceEmbeddingService
from label_sleuth.orchestrator.core.state_api.orchestrator_state_api import IterationStatus, OrchestratorStateApi
from label_sleuth.orchestrator.orchestrator_api import OrchestratorApi

print("user:")
print(getpass.getuser())

main_blueprint = Blueprint("main_blueprint", __name__)
executor = ThreadPoolExecutor(20)


class LabelSleuthApp(Flask):
    orchestrator_api: OrchestratorApi
    users: dict
    tokens: list


# in order for the IDE to recognize custom objects within the Label Sleuth flask application -- and specifically
# the methods of OrchestratorApi -- we define a class LabelSleuthApp with these objects and assign this type to
# the flask "current_app" proxy object
curr_app: LabelSleuthApp = current_app


def create_app(config: Configuration, output_dir) -> LabelSleuthApp:
    os.makedirs(output_dir, exist_ok=True)
    app = Flask(__name__, static_folder='./build')
    CORS(app)
    app.config['CORS_HEADERS'] = 'Content-Type'
    app.config["CONFIGURATION"] = config
    app.config["output_dir"] = output_dir
    app.users = {x['username']: dacite.from_dict(data_class=User, data=x) for x in app.config["CONFIGURATION"].users}
    app.tokens = [user.token for user in app.users.values()]
    sentence_embedding_service = SentenceEmbeddingService(embedding_model_dir=output_dir,
                                                          preload_spacy_model_name=config.language.spacy_model_name,
                                                          preload_fasttext_language_id=
                                                          config.language.fasttext_language_id)
    data_access = FileBasedDataAccess(output_dir)
    background_jobs_manager = BackgroundJobsManager()
    training_set_selection_factory = TrainingSetSelectionFactory(data_access, background_jobs_manager)

    app.orchestrator_api = OrchestratorApi(OrchestratorStateApi(os.path.join(output_dir, "workspaces")),
                                           data_access,
                                           ActiveLearningFactory(),
                                           ModelFactory(os.path.join(output_dir, "models"),
                                                        background_jobs_manager,
                                                        sentence_embedding_service),
                                           training_set_selection_factory,
                                           background_jobs_manager,
                                           sentence_embedding_service,
                                           app.config["CONFIGURATION"])
    app.register_blueprint(main_blueprint)
    return app


def start_server(app, host, port, num_serving_threads):
    disable_html_printouts = False
    if disable_html_printouts:
        logging.getLogger('werkzeug').disabled = True
        os.environ['WERKZEUG_RUN_MAIN'] = True

    from waitress import serve
    serve(app, host=host, port=port, threads=num_serving_threads)


@main_blueprint.route("/", defaults={'path': ''})
@main_blueprint.route('/<path:path>')  # catch all routes
def serve(path):
    if path != "" and os.path.exists(curr_app.static_folder + '/' + path):
        return send_from_directory(curr_app.static_folder, path)

    return send_from_directory(curr_app.static_folder, 'index.html')


@main_blueprint.route('/users/authenticate', methods=['POST'])
def login():
    post_data = request.get_json(force=True)
    username = post_data["username"]
    password = post_data["password"]

    if not verify_password(username, password):
        logging.warning(f"Login failed for username {username}")
        return make_response(jsonify({
            'error': "Login failed: wrong username or password"
        }), 401)
    else:
        user = curr_app.users.get(username)
        logging.info(f"LOGIN: {user.username}")
        return authenticate_response({
            'username': user.username,
            'token': user.token
        })


"""
Dataset endpoints. Datasets can be shared between workspaces, so these calls are not in the context of a particular
workspace.
"""


@main_blueprint.route("/datasets", methods=['GET'])
@login_if_required
def get_all_dataset_ids():
    """
    Get the names of all existing datasets
    """
    all_datasets = curr_app.orchestrator_api.get_all_dataset_names()
    res = {'datasets':
               [{"dataset_id": d} for d in all_datasets]}
    return jsonify(res)


@main_blueprint.route('/datasets/<dataset_name>/add_documents', methods=['POST'])
@cross_origin()
@login_if_required
def add_documents(dataset_name):
    """
    Upload a csv file, and add its contents as a collection of documents in the dataset *dataset_name*.
    If *dataset_name* does not already exist, it is created here. If it does exist, the process of adding documents also
    includes updating labels and model predictions for all workspaces that use this dataset.

    The uploaded csv file must follow the required format, using the relevant column names from DisplayFields.
    Specifically, the file must include a 'text' column, and may optionally include a 'document_id' column as well as
    metadata columns starting with the 'metadata_' prefix.

    :param dataset_name:
    """
    temp_dir = None
    try:
        csv_data = StringIO(request.files['file'].stream.read().decode("utf-8"))
        df = pd.read_csv(csv_data).rename(columns=lambda x: x.strip())
        temp_dir = os.path.join(curr_app.config["output_dir"], "temp", "csv_upload")
        temp_file_name = f"{next(tempfile._get_candidate_names())}.csv"
        os.makedirs(temp_dir, exist_ok=True)
        temp_file_path = os.path.join(temp_dir, temp_file_name)
        df.to_csv(os.path.join(temp_dir, temp_file_name))
        document_statistics, workspaces_to_update = \
            curr_app.orchestrator_api.add_documents_from_file(dataset_name, temp_file_path)
        return jsonify({"dataset_name": dataset_name,
                        "num_docs": document_statistics.documents_loaded,
                        "num_sentences": document_statistics.text_elements_loaded,
                        "workspaces_to_update": workspaces_to_update})
    except AlreadyExistsException as e:
        return jsonify({"dataset_name": dataset_name, "error": "documents already exist", "documents": e.documents,
                        "error_code": 409}), 409
    except Exception:
        logging.exception(f"failed to load or add documents to dataset '{dataset_name}'")
        return jsonify({"dataset_name": dataset_name, "error": traceback.format_exc(), "error_code": 400}), 400
    finally:
        if temp_dir is not None and os.path.exists(os.path.join(temp_dir, temp_file_name)):
            os.remove(os.path.join(temp_dir, temp_file_name))


@main_blueprint.route("/datasets/<dataset_name>/used_by", methods=['GET'])
@login_if_required
def get_workspaces_by_dataset_name(dataset_name):
    res = {'used_by': curr_app.orchestrator_api.get_workspaces_by_dataset_name(dataset_name)}
    return jsonify(res)


@main_blueprint.route("/datasets/<dataset_name>", methods=['DELETE'])
@login_if_required
def delete_dataset(dataset_name):
    """
    This call permanently deletes the given dataset. If the dataset is used by one or more workspaces, those will be deleted too, along with all the categories, user
    labels and models.

    :param dataset_name:
    """
    res = curr_app.orchestrator_api.delete_dataset(dataset_name)
    return jsonify(res)


"""
Workspace endpoints. Each workspace is associated with a particular dataset at creation time.
"""


@main_blueprint.route("/workspace", methods=['POST'])
@login_if_required
def create_workspace():
    """
    Create a new workspace

    :post_param workspace_id: str
    :post_param dataset_id: str
    """
    post_data = request.get_json(force=True)
    workspace_id = post_data["workspace_id"]
    dataset_name = post_data["dataset_id"]

    if curr_app.orchestrator_api.workspace_exists(workspace_id):
        logging.info(f"Trying to create workspace '{workspace_id}' which already exists")
        return jsonify({"workspace_id": workspace_id, "error": "Workspace already exists",
                        "error_code": 409}), 409
    curr_app.orchestrator_api.create_workspace(workspace_id=workspace_id, dataset_name=dataset_name)

    all_document_ids = curr_app.orchestrator_api.get_all_document_uris(workspace_id)
    first_document_id = all_document_ids[0]

    res = {'workspace': {'workspace_id': workspace_id,
                         'dataset_name': dataset_name,
                         'first_document_id': first_document_id}}

    return jsonify(res)


@main_blueprint.route("/workspaces", methods=['GET'])
@login_if_required
def get_all_workspace_ids():
    """
    Get the ids of all existing workspaces
    """
    res = {'workspaces': curr_app.orchestrator_api.list_workspaces()}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>", methods=['DELETE'])
@login_if_required
@validate_workspace_id
def delete_workspace(workspace_id):
    """
    This call permanently deletes all data associated with the given workspace, including all the categories, user
    labels and models.

    :param workspace_id:
    """
    curr_app.orchestrator_api.delete_workspace(workspace_id)
    return jsonify({'workspace_id': workspace_id})


@main_blueprint.route("/workspace/<workspace_id>", methods=['GET'])
@login_if_required
@validate_workspace_id
def get_workspace_info(workspace_id):
    """
    Get workspace information

    :param workspace_id:
    :returns a dictionary containing ids for the workspace and dataset as well as for all the documents in the dataset
    """
    document_ids = curr_app.orchestrator_api.get_all_document_uris(workspace_id)
    first_document_id = document_ids[0]

    res = {'workspace': {'workspace_id': workspace_id,
                         'dataset_name': curr_app.orchestrator_api.get_dataset_name(workspace_id),
                         'first_document_id': first_document_id,
                         'document_ids': document_ids}}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/load_dataset", methods=['POST'])
@login_if_required
@validate_workspace_id
def load_dataset(workspace_id):
    """
    Initial load of large datasets can take a while. The UI can call this endpoint to start a background loading of the
    dataset to the memory immediately after the workspace was selected. This will save time in subsequent calls to
    endpoints which require the dataset.

    :param workspace_id:
    """
    executor.submit(curr_app.orchestrator_api.preload_dataset, workspace_id)
    return jsonify({"success": True})


"""
Category endpoints. A category is defined in the context of a particular workspace. As a user works on the system, all
the labels, classification models etc. are associated with a specific category.
"""


@main_blueprint.route("/workspace/<workspace_id>/category", methods=['POST'])
@login_if_required
@validate_workspace_id
def create_category(workspace_id):
    """
    Create a new category in the given workspace

    :param workspace_id:
    :post_param category_name:
    :post_param category_description:
    """
    post_data = request.get_json(force=True)
    category_name = post_data["category_name"]
    category_description = post_data["category_description"]
    existing_category_names = [category.name for category
                               in curr_app.orchestrator_api.get_all_categories(workspace_id).values()]
    if category_name in existing_category_names:
        return jsonify({"workspace_id": workspace_id, "error": "A category with this name already exists",
                        "category_name": category_name, "error_code": 409}), 409

    category_id = curr_app.orchestrator_api.create_new_category(workspace_id, category_name, category_description)

    post_data['category_id'] = str(category_id)

    return jsonify(post_data)


@main_blueprint.route("/workspace/<workspace_id>/categories", methods=['GET'])
@login_if_required
@validate_workspace_id
def get_all_categories(workspace_id):
    """
    Get information about all existing categories in the workspace

    :param workspace_id:
    """
    categories = curr_app.orchestrator_api.get_all_categories(workspace_id)
    category_dicts = [{'category_id': id, 'category_name': category.name, 'category_description': category.description}
                      for id, category in categories.items()]

    res = {'categories': category_dicts}

    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/category/<category_id>", methods=['PUT'])
@login_if_required
@validate_workspace_id
def update_category(workspace_id, category_id):
    """
    :param workspace_id:
    :param category_id:
    :post_param category_name:
    :post_param category_description:
    """
    try:
        category_id = int(category_id)
    except:
        return jsonify({"type": "category_id_error",
                        "title": f"category_id should be an integer (got {category_id}) "}), 400

    if category_id not in curr_app.orchestrator_api.get_all_categories(workspace_id):
        return jsonify({"type": "category_id_does_not_exist",
                        "title": f"category_id {category_id} does not exist in workspace {workspace_id}"}), 404

    post_data = request.get_json(force=True)
    new_category_name = post_data["category_name"]
    new_category_description = post_data["category_description"]
    existing_category_names = [category.name for category
                               in curr_app.orchestrator_api.get_all_categories(workspace_id).values()]
    if new_category_name in existing_category_names:
        return jsonify({"workspace_id": workspace_id, "error": "A category with this name already exists",
                        "category_name": new_category_name, "error_code": 409}), 409

    curr_app.orchestrator_api.edit_category(workspace_id, category_id, new_category_name, new_category_description)
    return jsonify({"workspace_id": workspace_id, "category_id": str(category_id), "category_name": new_category_name,
                    "category_description": new_category_description})


@main_blueprint.route("/workspace/<workspace_id>/category/<category_id>", methods=['DELETE'])
@login_if_required
@validate_workspace_id
def delete_category(workspace_id, category_id):
    """
    This call permanently deletes all data associated with the given category, including user labels and models.

    :param workspace_id:
    :param category_id:
    """
    try:
        category_id = int(category_id)
    except:
        return jsonify({"type": "category_id_error",
                        "title": f"category_id should be an integer (got {category_id}) "}), 400

    if category_id not in curr_app.orchestrator_api.get_all_categories(workspace_id):
        return jsonify({"type": "category_id_does_not_exist",
                        "title": f"category_id {category_id} does not exist in workspace {workspace_id}"}), 404

    curr_app.orchestrator_api.delete_category(workspace_id, category_id)
    return jsonify({"workspace_id": workspace_id, 'category_id': str(category_id)})


"""
Document and element endpoints. The below calls are performed in the context of a particular workspace, and
optionally a particular category. Thus, they may return not just document and element information from the dataset, 
but also category labels, model predictions etc. that are associated with a particular document/element.
"""


@main_blueprint.route("/workspace/<workspace_id>/documents", methods=['GET'])
@login_if_required
@validate_workspace_id
def get_all_document_uris(workspace_id):
    """
    Get ids for all documents in the dataset associated with the workspace

    :param workspace_id:
    """

    doc_uris = curr_app.orchestrator_api.get_all_document_uris(workspace_id)
    res = {"documents": [{"document_id": uri} for uri in doc_uris]}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/document/<document_uri>", methods=['GET'])
@login_if_required
@validate_workspace_id
def get_document_elements(workspace_id, document_uri):
    """
    Get all elements in the given document

    :param workspace_id:
    :param document_uri:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """

    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].main_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
    document = curr_app.orchestrator_api.get_documents(workspace_id, dataset_name, [document_uri])[0]
    elements = document.text_elements
    category_id = request.args.get('category_id')
    if category_id is not None:
        category_id = int(category_id)
    hit_count = len(elements)
    elements = elements[start_idx: start_idx + size]
    elements_transformed = elements_back_to_front(workspace_id, elements, category_id)

    res = {'elements': elements_transformed, 'hit_count': hit_count}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/document/<document_uri>/positive_predictions", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_document_positive_predictions(workspace_id, document_uri):
    """
    Get elements in the given document that received a positive prediction from the relevant classification model,
    i.e. the latest trained model for the category specified in the request.

    :param workspace_id:
    :param document_uri:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    category_id = int(request.args['category_id'])
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))

    if len(curr_app.orchestrator_api.get_all_iterations_by_status(workspace_id, category_id,
                                                                  IterationStatus.READY)) == 0:
        elements_transformed = []
    else:
        dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
        document = curr_app.orchestrator_api.get_documents(workspace_id, dataset_name, [document_uri])[0]

        elements = document.text_elements

        predictions = [pred.label for pred in curr_app.orchestrator_api.infer(workspace_id, category_id, elements)]
        positive_predicted_elements = [element for element, prediction in zip(elements, predictions)
                                       if prediction == LABEL_POSITIVE]

        positive_predicted_elements = positive_predicted_elements[start_idx: start_idx + size]
        elements_transformed = elements_back_to_front(workspace_id, positive_predicted_elements, category_id)
    res = {'elements': elements_transformed}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/positive_predictions", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_positive_predictions(workspace_id):
    """
    Get elements in the given workspace that received a positive prediction from the relevant classification model,
    i.e. the latest trained model for the category specified in the request.
    As finding and returning _all_ the positively predicted elements is expensive, this endpoint only finds as many
    elements as is required by the request.

    :param workspace_id:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    category_id = int(request.args['category_id'])
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))

    all_ready_iterations = curr_app.orchestrator_api.get_all_iterations_by_status(workspace_id, category_id,
                                                                                  IterationStatus.READY)
    if len(all_ready_iterations) == 0:
        return jsonify({'elements': [], 'positive_fraction': None, 'total_count': None})
    else:
        positive_predicted_elements = curr_app.orchestrator_api.get_elements_by_prediction(
            workspace_id, category_id, LABEL_POSITIVE, sample_size=size, start_idx=start_idx, shuffle=False,
            remove_duplicates=False)  # For better performance in large datasets, we do not remove duplicates
        iteration, _ = all_ready_iterations[-1]

        elements_transformed = elements_back_to_front(workspace_id, positive_predicted_elements, category_id)

        positive_fraction = iteration.iteration_statistics["positive_fraction"]
        total_positive_count = iteration.iteration_statistics.get(
            'total_positive_count', int(positive_fraction * curr_app.orchestrator_api.get_text_element_count(workspace_id)))
        res = {'elements': elements_transformed,
               'positive_fraction': positive_fraction, 'total_count': total_positive_count}
        return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/element/<element_id>", methods=['GET'])
@login_if_required
@validate_workspace_id
def get_element_by_id(workspace_id, element_id):
    """
    Get the element with the given id

    :param workspace_id:
    :param element_id:
    :request_arg category_id:
    """
    category_id = request.args.get('category_id')
    if category_id is not None:
        category_id = int(category_id)
    return get_element(workspace_id, category_id, element_id)


@main_blueprint.route("/workspace/<workspace_id>/query", methods=['GET'])
@login_if_required
@validate_workspace_id
def query(workspace_id):
    """
    Query a dataset using the regular expression given in the request, and return elements that meet this search query

    :param workspace_id:
    :request_arg category_id:
    :request_arg qry_string: query string
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    category_id = request.args.get('category_id')
    if category_id is not None:
        category_id = int(category_id)
    query_string = request.args.get('qry_string')
    sample_size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    sample_start_idx = int(request.args.get('start_idx', 0))

    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
    resp = curr_app.orchestrator_api.query(workspace_id, dataset_name, category_id=None,
                                           query=query_string, is_regex=False,
                                           unlabeled_only=False, sample_size=sample_size,
                                           sample_start_idx=sample_start_idx, remove_duplicates=True)
    sorted_elements = sorted(resp["results"], key=lambda te: get_natural_sort_key(te.uri))
    elements_transformed = elements_back_to_front(workspace_id, sorted_elements, category_id)
    res = {'elements': elements_transformed,
           'hit_count': resp["hit_count"],
           'hit_count_unique': resp["hit_count_unique"]}
    return jsonify(res)


"""
Labeling-related endpoints.
"""


@main_blueprint.route('/workspace/<workspace_id>/element/<element_id>', methods=['PUT'])
@login_if_required
@validate_workspace_id
def set_element_label(workspace_id, element_id):
    """
    Update element label information. This endpoint either adds or removes a label for a given element and category,
    depending on the 'value' field in the post data.

    :param workspace_id:
    :param element_id:
    :post_param category_id:
    :post_param value: if value == 'none', this element's label for the given category will be removed. Otherwise, the
    element will be assigned a boolean label for the category corresponding to the given string (e.g., "true" -> True)
    :post_param update_counter: determines whether the label changes are reflected in the label change counters
    of the categories. Since an increase in label change counts can trigger the training of a new model, in some
    specific situations this parameter is set to False and the updating of the counter is performed at a later time.
    """
    post_data = request.get_json(force=True)

    if "category_id" not in post_data:
        return jsonify({"type": "missing_category_id", "title": "category_id was not provided in post data"}), 422

    category_id = int(post_data["category_id"])
    value = post_data["value"]
    update_counter = post_data.get('update_counter', True)

    if value == 'none':
        curr_app.orchestrator_api. \
            unset_labels(workspace_id, category_id, [element_id],
                         apply_to_duplicate_texts=curr_app.config["CONFIGURATION"].apply_labels_to_duplicate_texts)

    else:
        if value in ['true', "True", "TRUE", True]:
            value = True
        elif value in ['false', "False", "FALSE", False]:
            value = False
        else:
            raise Exception(f"cannot convert label to boolean. Input label = {value}")

        uri_with_updated_label = {element_id: {category_id: Label(value)}}
        curr_app.orchestrator_api. \
            set_labels(workspace_id, uri_with_updated_label,
                       apply_to_duplicate_texts=curr_app.config["CONFIGURATION"].apply_labels_to_duplicate_texts,
                       update_label_counter=update_counter)

    res = {'element': get_element(workspace_id, category_id, element_id), 'workspace_id': workspace_id,
           'category_id': str(category_id)}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/positive_elements", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_all_positively_labeled_elements_for_category(workspace_id):
    """
    Return all elements that were assigned a positive label for the given category by the user

    :param workspace_id:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))
    category_id = int(request.args['category_id'])
    elements, hit_count = get_all_labeled_elements(workspace_id, category_id, label=LABEL_POSITIVE, size=size,
                                                   start_idx=start_idx)

    res = {'elements': elements, "hit_count": hit_count}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/negative_elements", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_all_negatively_labeled_elements_for_category(workspace_id):
    """
    Return all elements that were assigned a negative label for the given category by the user

    :param workspace_id:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    category_id = int(request.args['category_id'])
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))
   
    elements, hit_count = get_all_labeled_elements(workspace_id, category_id, label=LABEL_NEGATIVE, size=size,
                                                   start_idx=start_idx)

    res = {'elements': elements, "hit_count": hit_count}
    return jsonify(res)


def get_all_labeled_elements(workspace_id, category_id, label=None, size: int = 100, start_idx: int = 0):
    """
    Return all elements that were assigned the given label, or all labeled elements if label is None

    :param workspace_id:
    :param category_id:
    :param label: return only elements assigned this label; if None, return all labeled elements
    :param size: number of elements to return
    :param start_idx: get elements starting from this index (for pagination)
    """
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
    elements = curr_app.orchestrator_api.get_all_labeled_text_elements(workspace_id, dataset_name, category_id,
                                                                       remove_duplicates=True)
    if label is not None:
        elements = [element for element in elements
                    if element.category_to_label[category_id].label == label]
    hit_count = len(elements)
    elements = elements[start_idx: start_idx + size]
    elements_transformed = elements_back_to_front(workspace_id, elements, category_id)
    return elements_transformed, hit_count


@main_blueprint.route('/workspace/<workspace_id>/import_labels', methods=['POST'])
@cross_origin()
@login_if_required
@validate_workspace_id
def import_labels(workspace_id):
    """
    Upload a csv file, and add its contents as user labels for the workspace. The file may contain labels for more than
    one category.
    The uploaded csv file must follow the required format, using the relevant column names from DisplayFields.
    Specifically, the file must include 'text', 'category_name' and 'label' columns, and may optionally include a
    'document_id' column.
    If document ids are provided AND app.config["CONFIGURATION"].apply_labels_to_duplicate_texts is False, these labels
    will be assigned only to elements in the specified documents; Otherwise, labels are assigned to all elements
    matching the given texts.

    :param workspace_id:
    """
    csv_data = StringIO(request.files['file'].stream.read().decode("utf-8"))
    df = pd.read_csv(csv_data).rename(columns=lambda x: x.strip())

    return jsonify(curr_app.orchestrator_api.import_category_labels(workspace_id, df))


@main_blueprint.route('/workspace/<workspace_id>/export_labels', methods=['GET'])
@login_if_required
@validate_workspace_id
def export_labels(workspace_id):
    """
    Download all user labels from the workspace as a csv file. Each row in the csv is a label for a specific element
    for a specific category. Column names for the various fields are listed under DisplayFields.

    :param workspace_id:
    :request_arg labeled_only: only export elements as they were labeled by the user. If set to False, use the
    TrainingSetSelectionStrategy to determine the exported elements
    """
    labeled_only = request.args.get('labeled_only', "true")
    labeled_only = labeled_only.lower() == "true"

    return curr_app.orchestrator_api.export_workspace_labels(workspace_id, labeled_only).to_csv(index=False)


"""
Models and Iterations
"""


@main_blueprint.route("/workspace/<workspace_id>/status", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_labelling_status(workspace_id):
    """
    Returns information about the number of user labels for the category, as well as a number between 0-100 that
    reflects when a new model will be trained.
    Once a certain amount of user labels for a given category has been reached, the get_labelling_status() call --
    via a call to orchestrator_api.train_if_recommended() -- will trigger an iteration flow in the background. This
    flow includes training a model, inferring the full corpus using this model, choosing candidate elements for
    labeling using active learning, and calculating various statistics.

    :param workspace_id:
    :request_arg category_id:
    """

    category_id = int(request.args['category_id'])
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)

    labeling_counts = curr_app.orchestrator_api. \
        get_label_counts(workspace_id, dataset_name, category_id,
                         remove_duplicates=curr_app.config["CONFIGURATION"].apply_labels_to_duplicate_texts)
    progress = curr_app.orchestrator_api.get_progress(workspace_id, dataset_name, category_id)

    # TODO move executor to the orchestrator
    executor.submit(curr_app.orchestrator_api.train_if_recommended, workspace_id, category_id)

    return jsonify({
        "labeling_counts": labeling_counts,
        "progress": progress
    })


@main_blueprint.route("/workspace/<workspace_id>/iterations", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
@cross_origin()
def get_all_iterations_for_category(workspace_id):
    """
    Return information about all the Iteration flows for this category and their current status.

    :param workspace_id:
    :request_arg category_id:
    :return: array of all iterations sorted from oldest to newest
    """
    category_id = int(request.args['category_id'])
    iterations = curr_app.orchestrator_api.get_all_iterations_for_category(workspace_id, category_id)
    res = {'iterations': extract_iteration_information_list(iterations)}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/active_learning", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_elements_to_label(workspace_id):
    """
    If at least one Iteration has completed for the given category, return a list of *size* elements that were
    recommended for labeling by the active learning module.

    :param workspace_id:
    :request_arg category_id:
    :request_arg size: the number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    category_id = int(request.args['category_id'])
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))

    if len(curr_app.orchestrator_api.get_all_iterations_by_status(workspace_id, category_id,
                                                                  IterationStatus.READY)) == 0:
        return jsonify({"elements": []})

    elements, hit_count = curr_app.orchestrator_api.get_elements_to_label(workspace_id, category_id, size, start_idx)
    elements_transformed = elements_back_to_front(workspace_id, elements, category_id)

    res = {'elements': elements_transformed, 'hit_count': hit_count}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/force_train", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def force_train_for_category(workspace_id):
    """
    This call is used for manually triggering a new Iteration flow.

    :param workspace_id:
    :request_arg category_id:
    """
    category_id = int(request.args['category_id'])
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)

    model_id = curr_app.orchestrator_api.train_if_recommended(workspace_id, category_id, force=True)

    labeling_counts = curr_app.orchestrator_api.get_label_counts(workspace_id, dataset_name, category_id)
    logging.info(f"force training a new model in workspace '{workspace_id}' for category '{category_id}', "
                 f"model id: {model_id}")

    return jsonify({
        "labeling_counts": labeling_counts,
        "model_id": model_id
    })


@main_blueprint.route('/workspace/<workspace_id>/export_predictions', methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def export_predictions(workspace_id):
    """
    Download all predictions of the model from iteration *iteration_index* of *category_id*, as a csv file.

    :param workspace_id:
    :request_arg category_id:
    :request_arg iteration_index: optional. if not provided, the model from the latest iteration will be exported.
    """
    category_id = int(request.args['category_id'])
    iteration_index = request.args.get('iteration_index', None)
    if iteration_index is None:
        _, iteration_index = curr_app.orchestrator_api. \
            get_all_iterations_by_status(workspace_id, category_id, IterationStatus.READY)[-1]
    else:
        iteration_index = int(iteration_index)
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
    elements = curr_app.orchestrator_api.get_all_text_elements(dataset_name)
    infer_results = curr_app.orchestrator_api.infer(workspace_id, category_id, elements,
                                                    iteration_index=iteration_index)
    return pd.DataFrame([{**te.__dict__, "score": pred.score, 'predicted_label': pred.label} for te, pred
                         in zip(elements, infer_results)]).to_csv(index=False)


@main_blueprint.route('/workspace/<workspace_id>/export_model', methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def export_model(workspace_id):
    """
    Download the trained model files for the given category and (optionally) iteration index.

    :param workspace_id:
    :request_arg category_id:
    :request_arg iteration_index: optional. if not provided, the model from the latest iteration will be exported.
    """

    usage_example = "from label_sleuth.models.util.standalone_inference import get_model_api\n\n" \
                    "model_path = \"<replace_with_the_exported_model_path>\"\n" \
                    "items_to_infer = [{\"text\": \"I love dogs\"}, {\"text\": \"I love cats\"}]\n" \
                    "model_api = get_model_api(model_path)\n" \
                    "model = model_api.load_model(model_path)\n" \
                    "predictions = model_api.infer(model, items_to_infer)\n" \
                    "for sentence_dict, pred in zip(items_to_infer, predictions):\n" \
                    "     print(f'sentence: \"{sentence_dict[\"text\"]}\" -> prediction: {pred}')\n"
    category_id = int(request.args['category_id'])
    iteration_index = request.args.get('iteration_index', None)
    logging.info(f"Exporting a model from workspace {workspace_id} category id {category_id}")
    if iteration_index is None:
        _, iteration_index = curr_app.orchestrator_api. \
            get_all_iterations_by_status(workspace_id, category_id, IterationStatus.READY)[-1]
    else:
        iteration_index = int(iteration_index)

    temp_model_dir = curr_app.orchestrator_api.copy_model_dir_for_export(workspace_id, category_id, iteration_index)
    try:
        memory_file = BytesIO()

        with zipfile.ZipFile(memory_file, 'w', compression=zipfile.ZIP_DEFLATED) as zf:
            for dirpath, dirnames, filenames in os.walk(temp_model_dir):
                for filename in filenames:
                    file_path = os.path.join(dirpath, filename)
                    archive_file_path = os.path.relpath(file_path, temp_model_dir)
                    zf.write(file_path, archive_file_path)

            # write usage example python file
            model_usage_example_file_path = os.path.join(temp_model_dir, "model_usage_example.py")
            with open(model_usage_example_file_path,'w') as text_file:
                text_file.write(usage_example)
            archive_file_path = os.path.relpath(model_usage_example_file_path, temp_model_dir)
            zf.write(model_usage_example_file_path, archive_file_path)

        memory_file.seek(0)
    finally:
        if os.path.exists(temp_model_dir):
            shutil.rmtree(temp_model_dir, ignore_errors=True)
    return send_file(memory_file, attachment_filename=f'model.zip', as_attachment=True)


"""
Labeling reports. These calls return labeled elements which the system suggests for review by the user, aiming to 
surface potential errors in the original labeling.
"""


@main_blueprint.route("/workspace/<workspace_id>/disagree_elements", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_label_and_model_disagreements(workspace_id):
    """
    Returns all labeled elements where the predictions of the latest model for the category differ from the label
    provided by the user.

    :param workspace_id:
    :request_arg category_id:
    """

    category_id = int(request.args['category_id'])
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
    elements = curr_app.orchestrator_api.get_all_labeled_text_elements(workspace_id, dataset_name, category_id)
    elements_transformed = elements_back_to_front(workspace_id, elements, category_id)

    res = {'disagree_elements':
               [element for element in elements_transformed
                if category_id in element['model_predictions'] and category_id in element['user_labels']
                and element['model_predictions'][category_id] != element['user_labels'][category_id]]}
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/suspicious_elements", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_suspicious_elements(workspace_id):
    """
    This call returns elements where the user label might be incorrect, based on an analysis of all elements
    labeled so far. Elements are ordered from most to least "suspicious" according to the analysis.

    The current implementation relies on the get_disagreements_using_cross_validation() function. In this
    implementation, several models are trained on different parts of the labeled data; labels are suspected as
    incorrect if a model's prediction on a left-out element disagrees with the user label for that element.

    :param workspace_id:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))
    category_id = int(request.args['category_id'])
    
    try:
        suspicious_elements = curr_app.orchestrator_api.get_suspicious_elements_report(workspace_id, category_id)
        hit_count = len(suspicious_elements)
        suspicious_elements = suspicious_elements[start_idx: start_idx + size]
        elements_transformed = elements_back_to_front(workspace_id, suspicious_elements, category_id)
        res = {'elements': elements_transformed, 'hit_count': hit_count}
        return jsonify(res)
    except Exception:
        logging.exception("Failed to generate suspicious elements report")
        res = {'elements': [], 'hit_count': 0}
        return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/contradiction_elements", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_contradicting_elements(workspace_id):
    """
    This call returns pairs of labeled elements that may be inconsistent with each other. Each pair consists of an
    element given a positive label and an element given a negative label; since the two elements are similar to each
    other, there is a possibility that one of the labels is incorrect. Pairs are sorted from most to least similar
    to each other.
    The current implementation of get_suspected_labeling_contradictions_by_distance() relies on distances between text
    embeddings to identify the opposite-label elements that are similar to each other.
    In addition, for each pair the call returns a list of unique tokens for each element, i.e. tokens that do not appear
    in the other element in the pair. This can be used to visualize the similarities/differences between the two texts.

    :param workspace_id:
    :request_arg category_id:
    :request_arg size: number of elements to return
    :request_arg start_idx: get elements starting from this index (for pagination)
    """
    category_id = int(request.args['category_id'])
    size = int(request.args.get('size', curr_app.config["CONFIGURATION"].sidebar_panel_elements_per_page))
    start_idx = int(request.args.get('start_idx', 0))
    try:
        contradiction_elements_dict = curr_app.orchestrator_api.get_contradiction_report(workspace_id, category_id)
        element_pairs_transformed = [elements_back_to_front(workspace_id, element_pair, category_id)
                                     for element_pair in contradiction_elements_dict['pairs']]
        hit_count = len(element_pairs_transformed)
        element_pairs_transformed = element_pairs_transformed[start_idx: start_idx + size]
        diffs = [[list(element_a_unique_token_set), list(element_b_unique_token_set)]
                 for element_a_unique_token_set, element_b_unique_token_set in contradiction_elements_dict['diffs']]
        res = {'pairs': element_pairs_transformed, 'diffs': diffs, 'hit_count': hit_count}
        return jsonify(res)
    except Exception:
        logging.exception(f"Failed to generate contradiction report for workspace "
                          f"{workspace_id} category_id {category_id}")
        return jsonify({"type": "contradiction_report_generation_error", "title":
            f"Failed to generate contradiction report for workspace {workspace_id} category_id {category_id}"}), 500


"""
Model evaluation
"""


@main_blueprint.route("/workspace/<workspace_id>/precision_evaluation_elements", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_elements_for_precision_evaluation(workspace_id):
    """
    Returns a sample of elements that were predicted as positive for the category by the latest model. This sample is
    used for evaluating the precision of the model predictions: the user is asked to label this sample of elements,
    and after they are labeled the run_precision_evaluation() call will use these labels to estimate model precision.

    :param workspace_id:
    :request_arg category_id:.
    """
    size = curr_app.config["CONFIGURATION"].precision_evaluation_size
    category_id = int(request.args['category_id'])
    random_state = len(curr_app.orchestrator_api.get_all_iterations_for_category(workspace_id, category_id))
    positive_predicted_elements = curr_app.orchestrator_api.\
        get_elements_by_prediction(workspace_id, category_id, required_prediction=LABEL_POSITIVE, sample_size=size,
                                   remove_duplicates=False, shuffle=True, random_state=random_state)
    elements_transformed = elements_back_to_front(workspace_id, positive_predicted_elements, category_id)
    logging.info(f"sampled {len(elements_transformed)} elements for evaluation")
    res = {'elements': elements_transformed}
    return jsonify(res)


@main_blueprint.route('/workspace/<workspace_id>/precision_evaluation_elements', methods=['POST'])
@login_if_required
@validate_category_id
@validate_workspace_id
@cross_origin()
def run_precision_evaluation(workspace_id):
    """
    Return a score estimating the precision of *model_id*. This method is triggered manually, and after the
    elements from get_elements_for_precision_evaluation() were labeled by the user.

    :param workspace_id:
    :request_arg category_id:
    :post_param iteration:
    :post_param ids: element ids of the elements labeled for the purpose of precision evaluation
    :post_param changed_elements_count: the number of labels that were added/changed in the labeling of elements
    for evaluation. In order to avoid triggering a new iteration before evaluation is complete, labels set in the
    context of precision evaluation are not immediately reflected in the label change counter of the category, and the
    counts are updated only when calling run_precision_evaluation() or cancel_precision_evaluation()
    """
    category_id = int(request.args['category_id'])
    post_data = request.get_json(force=True)
    ids = post_data["ids"]
    changed_elements_count = post_data["changed_elements_count"]
    iteration_index = post_data["iteration"]
    score = curr_app.orchestrator_api.estimate_precision(workspace_id, category_id, ids, changed_elements_count,
                                                         iteration_index)
    res = {'score': score}
    return jsonify(res)


@main_blueprint.route('/workspace/<workspace_id>/cancel_precision_evaluation', methods=['POST'])
@login_if_required
@validate_category_id
@validate_workspace_id
def cancel_precision_evaluation(workspace_id):
    """
    Exit the precision evaluation, and allow the evaluation elements labeled so far to be reflected in the
    label change counter

    :param workspace_id:
    :request_arg category_id:
    :post_param changed_elements_count: the number of labels that were added/changed in the labeling of elements
    for evaluation. In order to avoid triggering a new iteration before evaluation is complete, labels set in the
    context of precision evaluation are not immediately reflected in the label change counter of the category, and the
    counts are updated only when calling run_precision_evaluation() or cancel_precision_evaluation()

    """
    category_id = int(request.args['category_id'])
    post_data = request.get_json(force=True)
    changed_elements_count = post_data["changed_elements_count"]
    curr_app.orchestrator_api.increase_label_change_count_since_last_train(workspace_id, category_id,
                                                                           changed_elements_count)
    res = {'canceled': 'OK'}
    return jsonify(res)



"""
Information on enriched tokens in a specific subset of elements, used for visualization.
"""


@main_blueprint.route("/workspace/<workspace_id>/labeled_info_gain", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
def get_labeled_elements_enriched_tokens(workspace_id):
    """
    Returns tokens and bigrams that are characteristic of positively labeled elements versus negatively labeled
    elements for the category, along with weights indicating the relative significance of each token/bigram. These are
    used to visualize the characteristic vocabulary of the category, e.g. using a word cloud. The current
    implementation relies on calculating Mutual Information between the ngrams and the user labels.

    :param workspace_id:
    :request_arg category_id:
    """
    category_id = int(request.args['category_id'])
    dataset_name = curr_app.orchestrator_api.get_dataset_name(workspace_id)
    elements = curr_app.orchestrator_api.get_all_labeled_text_elements(workspace_id, dataset_name, category_id)
    res = dict()
    if elements and len(elements) > 0:
        boolean_labels = [element.category_to_label[category_id].label == LABEL_POSITIVE for element in elements]
        res['info_gain'] = extract_enriched_ngrams_and_weights_list(elements, boolean_labels)
    return jsonify(res)


@main_blueprint.route("/workspace/<workspace_id>/predictions_info_gain", methods=['GET'])
@login_if_required
@validate_category_id
@validate_workspace_id
@cross_origin()
def get_predictions_enriched_tokens(workspace_id):
    """
    Returns tokens and bigrams that are characteristic of positively predicted elements versus negatively predicted
    elements by the latest model for the category, along with weights indicating the relative significance of each
    token/bigram. These are used to visualize the characteristic vocabulary of the model predictions, e.g. using a
    word cloud. The current implementation relies on calculating Mutual Information between the ngrams and the model
    predictions.

    :param workspace_id:
    :request_arg category_id:
    """
    res = dict()

    category_id = int(request.args['category_id'])
    if len(curr_app.orchestrator_api.get_all_iterations_by_status(workspace_id, category_id,
                                                                  IterationStatus.READY)) == 0:
        res['info_gain'] = []
        return jsonify(res)

    true_elements = curr_app.orchestrator_api.get_elements_by_prediction(
        workspace_id, category_id, LABEL_POSITIVE, sample_size=1000, shuffle=True, remove_duplicates=False)
    false_elements = curr_app.orchestrator_api.get_elements_by_prediction(
        workspace_id, category_id, LABEL_NEGATIVE, sample_size=1000, shuffle=True, remove_duplicates=False)

    elements = true_elements + false_elements
    boolean_labels = [LABEL_POSITIVE] * len(true_elements) + [LABEL_NEGATIVE] * len(false_elements)

    res['info_gain'] = extract_enriched_ngrams_and_weights_list(elements, boolean_labels)
    return jsonify(res)


@main_blueprint.route("/feature_flags", methods=['GET'])
def get_feature_flags():
    """Returns the value of the feature flags

    Feature flags are a subset of config["CONFIGURATION"].
    """

    res =  {
        "login_required": curr_app.config['CONFIGURATION'].login_required,
        "main_panel_elements_per_page": curr_app.config['CONFIGURATION'].main_panel_elements_per_page,
        "sidebar_panel_elements_per_page": curr_app.config['CONFIGURATION'].sidebar_panel_elements_per_page,
        "right_to_left": curr_app.config['CONFIGURATION'].language.right_to_left
    }
    logging.debug(f'Feature flags are: {res}')
    return jsonify(res) 
    

@main_blueprint.route("/version", methods=['GET'])
def get_git_describe():
    """Retrieves the system version 

    Git version is given priority, thus if git is installed 
    in the system and a .git folder is present, the output
    of 'git describe' is returned. If that fails, the package
    version is returned. If that fails too, the string 'not
    available' is returned instead.

    Returns:
        an object with version and source fields
    """
    try:
        import git
        repo = git.repo.Repo(os.path.abspath(os.path.join(__file__, os.pardir, os.pardir)))
        version = repo.git.describe(tags=True)
        source = 'git'
    except:
        try:
            version = pkg_resources.require("label-sleuth")[0].version
            source = 'pypi'
        except:
            version = 'Not available'
            source = None
            logging.warning(f'Could not get Label Sleuth version information')

    return {'version': version, 'source': source}


#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import re
from typing import Set

import pandas as pd
from label_sleuth.data_access.core.data_structs import TextElement, URI_SEP, LabelType
from label_sleuth.data_access.data_access_api import LabeledStatus


def get_dataset_name_from_uri(uri):
    uri_split = uri.split(URI_SEP)
    return uri_split[0]


def get_sort_key_by_document_name(uri):
    def natural_sort(text):
        return [int(x) if x.isdigit() else x for x in re.split(r'(\d+)', text)]

    doc_name = uri.split(URI_SEP)[1]
    return natural_sort(doc_name)


def uri_to_filename(uri):
    filename = uri.rstrip(URI_SEP)
    return filename


def filename_to_uri(filename):
    uri = filename
    return uri


def build_text_elements_from_dataframe_and_labels(df, labels_dict):
    # text element fields are extracted from the dataframe, with the exception of the labels, which are stored elsewhere
    element_data_columns = list(TextElement.get_field_names() - {'category_to_label'})
    element_dicts = map(lambda row: dict(zip(element_data_columns, row)), df[element_data_columns].values)
    text_elements = [TextElement(**d, category_to_label=labels_dict.get(d['uri'], {}).copy())
                     for d in element_dicts]

    return text_elements


def filter_by_labeled_status(df: pd.DataFrame, labels_series: pd.Series, category_id: int, labeled_status: LabeledStatus,
                             label_types: Set[LabelType] = None):
    """
    :param df:
    :param labels_series: series of the label dictionary corresponding to each element in the dataframe
    :param category_id:
    :param labeled_status: unlabeled, labeled or all
    :param label_types: set of applicable label types if filtering for labeled elements (LabelStatus.LABELED)
    :return:
    """
    if labeled_status in [LabeledStatus.UNLABELED, LabeledStatus.ALL] and label_types is not None:
        raise Exception(f"Label type is inapplicable when fetching {labeled_status} elements")

    if labeled_status == LabeledStatus.LABELED and label_types is None:
        raise Exception(f"label_types must be provided when filtering labeled elements")

    if labeled_status == LabeledStatus.UNLABELED:
        return df[labels_series.apply(lambda x: category_id not in x)]
    elif labeled_status == LabeledStatus.LABELED:
        return df[labels_series.apply(lambda x: category_id in x and x[category_id].label_type in label_types)]

    return df


def filter_by_query_and_document_uri(df: pd.DataFrame, query, is_regex: bool = False, document_id=None):
    if document_id is not None:
        df = df[df.uri.str.startswith(f"{document_id}-")]
    if query:
        # case=is_regex: we want the query to be case sensitive if we are matching using a regex and case insensitive otherwise
        df = df[df.text.str.contains(query, case=is_regex, na=False, regex=is_regex)]
    return df


def filter_by_query_and_label_status(df: pd.DataFrame, labels_series: pd.Series, category_id: int,
                                     labeled_status: LabeledStatus, query: str, is_regex: bool = False,
                                     label_types: Set[LabelType] = None):
    """
    :param df:
    :param labels_series: series of the label dictionary corresponding to each element in the dataframe
    :param category_id:
    :param labeled_status: unlabeled, labeled or all
    :param query: query to use for filtering text elements
    :param is_regex: whether to process the query as regular expression
    :param label_types: set of applicable label types if filtering for labeled elements (LabelStatus.LABELED)
    :return:
    """
    df = filter_by_labeled_status(df, labels_series, category_id, labeled_status, label_types=label_types)
    return filter_by_query_and_document_uri(df, query, is_regex)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import random
import unittest
from collections import Counter
from typing import List
import tempfile
from label_sleuth.data_access.core.data_structs import Document, TextElement, Label
from label_sleuth.data_access.file_based.utils import URI_SEP

from label_sleuth.data_access.file_based.file_based_data_access import FileBasedDataAccess
from label_sleuth.data_access.core.data_structs import LABEL_POSITIVE, LABEL_NEGATIVE


def generate_simple_doc(dataset_name, doc_id=0, add_duplicate=False):
    sentences = [f'Document Title is Super Interesting {doc_id}', f'First sentence is not that attractive. {doc_id}',
                 f'The second one is a bit better. {doc_id}',
                 f'Last sentence offers a promising view for the future! {doc_id}']
    if add_duplicate:
        sentences.append(sentences[0])
    text_elements = []
    start_span = 0
    for idx, sentence in enumerate(sentences):
        end_span = start_span + len(sentence)
        text_elements.append(TextElement(uri=URI_SEP.join([dataset_name, str(doc_id), str(idx)]), text=sentence,
                                         span=[(start_span, end_span)], metadata={}, category_to_label={}))
        start_span = end_span + 1

    doc = Document(uri=dataset_name + URI_SEP + str(doc_id), text_elements=text_elements, metadata={})
    return doc


def generate_corpus(data_access, dataset_name, num_of_documents=1, add_duplicate=False):
    data_access.delete_dataset(dataset_name)
    docs = [generate_simple_doc(dataset_name, doc_id, add_duplicate) for doc_id in range(0, num_of_documents)]
    data_access.add_documents(dataset_name=dataset_name, documents=docs)
    return docs


def add_random_labels_to_document(doc: Document, min_num_sentences_to_label: int, categories: List[int], seed=0):
    uri_to_label = {}
    random.seed(seed)
    text_elements_to_label = random.sample(doc.text_elements, min(min_num_sentences_to_label, len(doc.text_elements)))
    for elem in text_elements_to_label:
        categories_to_label = random.sample(categories, random.randint(0, len(categories)))
        labels = {cat: Label(label=LABEL_POSITIVE) if cat in categories_to_label else Label(
            label=LABEL_NEGATIVE) for cat in categories}
        uri_to_label[elem.uri] = labels
    return uri_to_label


def add_labels_to_doc(doc: Document, category_id: int):
    sentences_and_labels = []
    for elem in doc.text_elements:
        labels = {category_id: Label(label=LABEL_POSITIVE)}
        sentences_and_labels.append((elem.uri, labels))
    return sentences_and_labels


class TestFileBasedDataAccess(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.data_temp_dir = tempfile.TemporaryDirectory()
        cls.data_access = FileBasedDataAccess(cls.data_temp_dir.name)

    @classmethod
    def tearDownClass(cls):
        cls.data_temp_dir.cleanup()

    def test_add_documents_and_get_documents(self):
        dataset_name = self.test_add_documents_and_get_documents.__name__ + '_dump'
        doc = generate_corpus(self.data_access, dataset_name)[0]
        doc_in_memory = self.data_access.get_documents(None, dataset_name, [doc.uri])[0]
        # compare all fields
        diffs = [(field, getattr(doc_in_memory, field), getattr(doc, field)) for field in
                 Document.__annotations__ if not getattr(doc_in_memory, field) == getattr(doc, field)]
        self.assertEqual(0, len(diffs))
        self.data_access.delete_dataset(dataset_name)

    def test_set_labels_and_get_documents(self):
        workspace_id = 'test_set_labels'
        dataset_name = self.test_set_labels_and_get_documents.__name__ + '_dump'
        categories = ['cat_' + str(i) for i in range(3)]
        doc = generate_corpus(self.data_access, dataset_name)[0]
        uri_to_label = add_random_labels_to_document(doc, 5, categories)
        self.data_access.set_labels(workspace_id, uri_to_label)

        doc_with_labels_info = self.data_access.get_documents(workspace_id, dataset_name, [doc.uri])
        texts_and_labels_dict = dict(uri_to_label)
        for text in doc_with_labels_info[0].text_elements:
            if text.uri in texts_and_labels_dict:
                self.assertDictEqual(text.category_to_label, texts_and_labels_dict[text.uri])
            else:
                self.assertDictEqual(text.category_to_label, {})
        self.data_access.delete_dataset(dataset_name)

    def test_unset_labels(self):
        workspace_id = 'test_unset_labels'
        dataset_name = self.test_set_labels_and_get_documents.__name__ + '_dump'
        category_id = 0
        doc = generate_corpus(self.data_access, dataset_name)[0]
        texts_and_labels_list = add_labels_to_doc(doc, category_id)
        self.data_access.set_labels(workspace_id, dict(texts_and_labels_list))

        labels_count = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        self.assertGreater(labels_count['true'], 0)
        self.data_access.unset_labels(workspace_id, category_id, [x[0] for x in texts_and_labels_list])
        labels_count_after_unset = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        self.assertEqual(0, labels_count_after_unset['true'])
        self.data_access.delete_dataset(dataset_name)

    def test_get_all_document_uris(self):
        dataset_name = self.test_get_all_document_uris.__name__ + '_dump'
        docs = generate_corpus(self.data_access, dataset_name, random.randint(1, 10))
        docs_uris_in_memory = self.data_access.get_all_document_uris(dataset_name)
        docs_uris_expected = [doc.uri for doc in docs]
        self.assertSetEqual(set(docs_uris_expected), set(docs_uris_in_memory))
        self.data_access.delete_dataset(dataset_name)

    def test_get_all_text_elements_uris(self):
        dataset_name = self.test_get_all_text_elements_uris.__name__ + '_dump'
        docs = generate_corpus(self.data_access, dataset_name, random.randint(1, 10))
        text_elements_uris_in_memory = self.data_access.get_all_text_elements_uris(dataset_name)
        text_elements_uris_expected = [text.uri for doc in docs for text in doc.text_elements]
        self.assertSetEqual(set(text_elements_uris_expected), set(text_elements_uris_in_memory))
        self.data_access.delete_dataset(dataset_name)

    def test_get_all_text_elements(self):
        dataset_name = self.test_get_all_text_elements.__name__ + '_dump'
        docs = generate_corpus(self.data_access, dataset_name, random.randint(1, 10))
        text_elements_found = self.data_access.get_all_text_elements(dataset_name)
        text_elements_found.sort(key=lambda t: t.uri)
        text_elements_expected = [text for doc in docs for text in doc.text_elements]
        text_elements_expected.sort(key=lambda t: t.uri)
        self.assertListEqual(text_elements_expected, text_elements_found)

        del self.data_access.ds_in_memory[dataset_name]
        # test again after clearing from memory
        text_elements_found = self.data_access.get_all_text_elements(dataset_name)
        text_elements_found.sort(key=lambda t: t.uri)
        self.assertListEqual(text_elements_expected, text_elements_found)
        self.data_access.delete_dataset(dataset_name)

    # def test_get_text_elements(self):
    #     dataset_name = self.test_sample_text_elements.__name__ + '_dump'
    #     sample_size = 5
    #     generate_corpus(self.data_access,dataset_name, 10)
    #     sampled_texts_res = self.data_access.get_text_elements(dataset_name, sample_size)
    #     self.assertEqual(sample_size, len(sampled_texts_res['results']))
    #
    #     sample_all = 10 ** 100  # a huge sample_size to sample all elements
    #     sampled_texts_res = self.data_access.get_text_elements(dataset_name, sample_all)
    #     self.assertEqual(sampled_texts_res['hit_count'], len(sampled_texts_res['results']),
    #                      f'the number of sampled elements does not equal to the hit count, '
    #                      f'even though asked to sample all.')
    #     self.assertEqual(len(self.data_access.get_all_text_elements_uris(dataset_name)), sampled_texts_res['hit_count'],
    #                      f'the hit count does not equal to the total number of element uris in the dataset, '
    #                      f'even though asked to sample all.')
    #     # assert no labels were added
    #     self.assertDictEqual(sampled_texts_res['results'][0].category_to_label, {})
    #     self.data_access.delete_dataset(dataset_name)

    def test_get_text_elements(self):
        def sample_and_check_that_labels_match(doc, orig_dict):
            sampled_texts_res = self.data_access.get_text_elements(workspace_id, dataset_name, sample_all)
            for doc_text in doc.text_elements:
                sampled_text = [sampled for sampled in sampled_texts_res['results'] if sampled.uri == doc_text.uri]
                self.assertEqual(1, len(sampled_text))
                if sampled_text[0].uri in orig_dict:
                    self.assertDictEqual(sampled_text[0].category_to_label, orig_dict[sampled_text[0].uri],
                                         f'for text {doc_text}')
                else:
                    self.assertDictEqual(sampled_text[0].category_to_label, {}, f'for text {doc_text}')

        workspace_id = 'test_sample_text_elements'
        dataset_name = self.test_get_text_elements.__name__ + '_dump'
        sample_all = 10 ** 100  # a huge sample_size to sample all elements
        docs = generate_corpus(self.data_access, dataset_name, 5)
        # add labels info for a single doc
        selected_doc = docs[0]
        texts_and_labels_dict = dict(add_random_labels_to_document(selected_doc, 3, [0, 1]))
        self.data_access.set_labels(workspace_id, texts_and_labels_dict)

        sample_and_check_that_labels_match(selected_doc, texts_and_labels_dict)
        del self.data_access.ds_in_memory[dataset_name]
        # check again after clearing from memory
        sample_and_check_that_labels_match(selected_doc, texts_and_labels_dict)
        self.data_access.delete_dataset(dataset_name)

    def test_get_unlabeled_text_elements(self):
        workspace_id = 'test_sample_unlabeled_text_elements'
        dataset_name = self.test_get_unlabeled_text_elements.__name__ + '_dump'
        category_id = 0
        sample_all = 10 ** 100  # a huge sample_size to sample all elements
        docs = generate_corpus(self.data_access, dataset_name, 2)
        # add labels info for a single doc
        selected_doc = docs[0]
        texts_and_labels_list = add_random_labels_to_document(selected_doc, 5, [category_id])
        self.data_access.set_labels(workspace_id, texts_and_labels_list)

        sampled_texts_res = self.data_access.get_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                                         sample_all)
        for sampled_text in sampled_texts_res['results']:
            self.assertDictEqual(sampled_text.category_to_label, {})
        self.data_access.delete_dataset(dataset_name)

    def test_get_labeled_text_elements(self):
        workspace_id = 'test_sample_labeled_text_elements'
        dataset_name = self.test_get_labeled_text_elements.__name__ + '_dump'
        category_id = 2
        sample_all = 10 ** 100  # a huge sample_size to sample all elements
        docs = generate_corpus(self.data_access, dataset_name, 2)
        # add labels info for a single doc
        selected_doc = docs[0]
        texts_and_labels_list = add_random_labels_to_document(selected_doc, 5, [category_id])
        self.data_access.set_labels(workspace_id, texts_and_labels_list)
        texts_and_labels_dict = dict(texts_and_labels_list)

        sampled_texts_res = self.data_access.get_labeled_text_elements(workspace_id, dataset_name, category_id,
                                                                       sample_all)
        self.assertEqual(len(texts_and_labels_list), len(sampled_texts_res['results']),
                         f'all and only the {len(texts_and_labels_list)} labeled elements should have been sampled.')

        for sampled_text in sampled_texts_res['results']:
            self.assertIn(sampled_text.uri, texts_and_labels_dict.keys(),
                          f'the sampled text uri - {sampled_text.uri} - was not found in the '
                          f'texts that were labeled: {texts_and_labels_dict}')
            self.assertDictEqual(sampled_text.category_to_label, texts_and_labels_dict[sampled_text.uri])
        self.data_access.delete_dataset(dataset_name)

    def test_sample_by_query_text_elements(self):
        workspace_id = 'test_sample_by_query_text_elements'
        dataset_name = self.test_sample_by_query_text_elements.__name__ + '_dump'
        category_id = 0
        query = 'sentence'
        sample_all = 10 ** 100  # a huge sample_size to sample all elements
        doc = generate_corpus(self.data_access, dataset_name, 1)[0]
        # doc's elements = ['Document Title is Super Interesting', 'First sentence is not that attractive.',
        #          'The second one is a bit better.', 'Last sentence offers a promising view for the future!']
        # add labels info for a single doc
        uri_to_label_dict = {doc.text_elements[0].uri: {category_id: Label(label=LABEL_POSITIVE)},
                             # 1st sent does not match query
                             doc.text_elements[1].uri: {
                                 category_id: Label(label=LABEL_POSITIVE)}}  # 2nd sent does match query

        self.data_access.set_labels(workspace_id, uri_to_label_dict)

        # query + unlabeled elements
        sampled_texts_res = self.data_access.get_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                                         sample_all)
        for sampled_text in sampled_texts_res['results']:
            self.assertDictEqual(sampled_text.category_to_label, {})

        # query + labeled elements
        sampled_texts_res = self.data_access.get_labeled_text_elements(workspace_id, dataset_name, category_id, sample_all,
                                                                       query)
        self.assertEqual(1, len(sampled_texts_res['results']),
                         f'all and only the {len(uri_to_label_dict)} labeled elements should have been sampled.')

        for sampled_text in sampled_texts_res['results']:
            self.assertIn(sampled_text.uri, uri_to_label_dict.keys(),
                          f'the sampled text uri - {sampled_text.uri} - was not found in the '
                          f'texts that were labeled: {uri_to_label_dict}')
            self.assertIn(query, sampled_text.text)
        self.data_access.delete_dataset(dataset_name)

    def test_query_text_elements_pagination(self):
        workspace_id = 'test_sample_by_query_text_elements'
        dataset_name = self.test_sample_by_query_text_elements.__name__ + '_dump'
        category_id = 0
        query = 'sentence'
        sample_all = 10 ** 100  # a huge sample_size to sample all elements
        generate_corpus(self.data_access, dataset_name, 5)

        # query no limit
        sampled_texts_res = self.data_access.get_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                                         sample_all)
        self.assertEqual(20, len(sampled_texts_res['results']),
                         f"20 elements should have been sampled.")

        second_element_uri = sampled_texts_res["results"][1].uri
        sampled_texts_res = self.data_access.get_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                                         sample_all, sample_start_idx=1)
        self.assertEqual(19, len(sampled_texts_res['results']),
                         f'19 elements should have been sampled.')

        self.assertEqual(second_element_uri, sampled_texts_res['results'][0].uri,
                         "when using sample_start_idx=1, the first element should be the second element without paging")

        sampled_texts_res = self.data_access.get_unlabeled_text_elements(workspace_id, dataset_name, category_id,
                                                                         sample_all, sample_start_idx=20)
        self.assertEqual(0, len(sampled_texts_res['results']),
                         "when sample_start_idx=number_of_elements, no element should return")

        self.data_access.delete_all_labels(workspace_id, dataset_name)
        self.data_access.delete_dataset(dataset_name)
        # texts_and_labels_dict = dict(texts_and_labels_list)
        # for sampled_text in sampled_texts_res['results']:
        #     self.assertIn(sampled_text.uri, texts_and_labels_dict.keys(),
        #                   f'the sampled text uri - {sampled_text.uri} - was not found in the '
        #                   f'texts that were labeled: {texts_and_labels_dict}')
        #     self.assertIn(query, sampled_text.text)
        # self.data_access.delete_dataset(dataset_name)

    def test_multithread_async_set_labels(self):
        import threading
        num_docs = 100
        workspace_id = 'test_multithread_set_labels'
        dataset_name = self.test_multithread_async_set_labels.__name__ + '_dump'
        category_id = 0
        categories = [category_id]
        corpus = generate_corpus(self.data_access, dataset_name, num_of_documents=num_docs)
        batch_size = 10

        threads = []
        results = [None] * int(num_docs / batch_size)
        i = 0
        for begin in range(0, num_docs, batch_size):
            end = min(begin + batch_size, num_docs)
            x = threading.Thread(target=self.set_labels_thread, args=(categories, category_id, corpus,
                                                                      workspace_id, begin, end, i, results))
            i += 1
            threads.append(x)
            x.start()
        for thread in threads:
            thread.join()
        all_from_threads = Counter()
        for d in results:
            all_from_threads.update(d)
        self.assertDictEqual(dict(all_from_threads), dict(
            self.data_access.get_label_counts(workspace_id, dataset_name, category_id)))
        self.data_access.delete_dataset(dataset_name)

    def set_labels_thread(self, categories, category_id, corpus, workspace_id, start, stop, idx, results):
        sentences_and_labels_all = {}
        for doc in corpus[start:stop]:
            texts_and_labels_list = add_random_labels_to_document(doc, 5, categories)
            self.data_access.set_labels(workspace_id, texts_and_labels_list)
            sentences_and_labels_all.update(texts_and_labels_list)
        true_count = len(
            [1 for _, label_dict in sentences_and_labels_all.items() if label_dict[category_id].label == LABEL_POSITIVE])
        false_count = len(
            [1 for _, label_dict in sentences_and_labels_all.items() if label_dict[category_id].label == LABEL_NEGATIVE])
        results[idx] = {str(LABEL_POSITIVE).lower(): true_count, str(LABEL_NEGATIVE).lower(): false_count}

    def test_get_label_counts(self):
        workspace_id = 'test_get_label_counts'
        dataset_name = self.test_get_label_counts.__name__ + '_dump'
        category_to_count = 2
        docs = generate_corpus(self.data_access, dataset_name, 2)
        # add labels info for a single doc
        selected_doc = docs[0]
        uri_to_label_dict = add_random_labels_to_document(selected_doc, 5, [category_to_count, 0, 1])
        # if texts_and_labels_list:
        #     if category in texts_and_labels_list[0][1]:
        #         texts_and_labels_list[0][1][category].label = LABEL_NEGATIVE
        #     else:
        #         texts_and_labels_list[0][1][category] = Label(label=LABEL_NEGATIVE)
        self.data_access.set_labels(workspace_id, uri_to_label_dict)

        category_label_counts = self.data_access.get_label_counts(workspace_id, dataset_name, category_to_count)
        for label_val, observed_count in category_label_counts.items():
            expected_count = len(
                [label for uri, label in uri_to_label_dict.items() if
                 category_to_count in label and label_val == str(label[category_to_count].label).lower()])
            self.assertEqual(expected_count, observed_count, f'count for {label_val} does not match.')
        self.data_access.delete_dataset(dataset_name)

    def test_get_text_elements_by_id(self):
        workspace_id = "test_get_text_elements_by_id"
        dataset_name = self.test_get_text_elements_by_id.__name__ + '_dump'
        categories = [i for i in range(3)]
        docs = generate_corpus(self.data_access, dataset_name, 2)
        doc = docs[0]
        texts_and_labels_list = add_random_labels_to_document(doc, 5, categories)  # [(uri, {category: Label})]
        uri_to_labels = dict(texts_and_labels_list)
        self.data_access.set_labels(workspace_id, texts_and_labels_list)
        uris = [x.uri for doc in docs for x in doc.text_elements]
        uris = [uris[2], uris[0], uris[1]]  # using different order to test if order is preserved

        all_elements = self.data_access.get_text_elements_by_uris(workspace_id, dataset_name, uris)
        self.assertEqual(len(uris), len(all_elements))
        self.assertEqual(uri_to_labels[uris[0]], all_elements[0].category_to_label)
        self.assertEqual(uri_to_labels[uris[1]], all_elements[1].category_to_label)

        del self.data_access.ds_in_memory[dataset_name]
        # test again after clearing from memory
        all_elements = self.data_access.get_text_elements_by_uris(workspace_id, dataset_name, uris)
        self.assertEqual(len(uris), len(all_elements))
        self.assertEqual(uri_to_labels[uris[0]], all_elements[0].category_to_label)
        self.assertEqual(uri_to_labels[uris[1]], all_elements[1].category_to_label)

        self.data_access.delete_dataset(dataset_name)

    def test_duplicates_removal(self):
        workspace_id = 'test_duplicates_removal'
        dataset_name = self.test_duplicates_removal.__name__ + '_dump'
        generate_corpus(self.data_access, dataset_name, 1, add_duplicate=True)
        category_id = 0
        all_elements = sorted(self.data_access.get_all_text_elements(dataset_name),
                              key=lambda te: te.uri)
        all_elements2 = sorted(
            self.data_access.get_text_elements(workspace_id, dataset_name, 10 ** 6, remove_duplicates=False)['results'],
            key=lambda te: te.uri)
        self.assertListEqual(all_elements, all_elements2)
        all_without_dups = self.data_access.get_text_elements(workspace_id, dataset_name, 10 ** 6,
                                                              remove_duplicates=True)['results']
        self.assertEqual(len(all_elements), len(all_without_dups) + 1)

        # 1. test propagation of labels:
        uri_to_label_dict = {elem.uri: {category_id: Label(label=LABEL_POSITIVE)}
                             for elem in all_without_dups}
        # set labels without propagating to duplicates
        self.data_access.set_labels(workspace_id, uri_to_label_dict, apply_to_duplicate_texts=False)
        labels_count = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        self.assertEqual(labels_count[str(LABEL_POSITIVE).lower()], len(all_without_dups))
        # unset labels
        self.data_access.unset_labels(workspace_id, category_id, [elem.uri for elem in all_without_dups])
        labels_count = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        self.assertEqual(labels_count[str(LABEL_POSITIVE).lower()], 0)
        # set labels with propagating to duplicates
        self.data_access.set_labels(workspace_id, uri_to_label_dict, apply_to_duplicate_texts=True)
        labels_count = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        self.assertEqual(labels_count[str(LABEL_POSITIVE).lower()], len(all_elements))
        self.data_access.unset_labels(workspace_id, category_id, [elem.uri for elem in all_elements])

        # 2. test sampling of duplicate examples:
        non_representative_duplicates = [x for x in all_elements if x not in all_without_dups]
        uri_to_label_dict = {elem.uri: {category_id: Label(label=LABEL_POSITIVE)}
                             for elem in non_representative_duplicates}
        # set labels without propagating to duplicates
        self.data_access.set_labels(workspace_id, uri_to_label_dict, apply_to_duplicate_texts=False)
        labels_count = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        sampled = \
            self.data_access.get_labeled_text_elements(workspace_id, dataset_name, category_id, 10 ** 6,
                                                       remove_duplicates=True)['results']
        self.assertEqual(labels_count[str(LABEL_POSITIVE).lower()], len(sampled), len(non_representative_duplicates))
        # set labels with propagating to duplicates
        self.data_access.set_labels(workspace_id, uri_to_label_dict, apply_to_duplicate_texts=True)
        labels_count = self.data_access.get_label_counts(workspace_id, dataset_name, category_id)
        sampled = \
            self.data_access.get_labeled_text_elements(workspace_id, dataset_name, category_id, 10 ** 6,
                                                       remove_duplicates=True)['results']
        self.assertGreater(labels_count[str(LABEL_POSITIVE).lower()], len(non_representative_duplicates))
        labels_count_no_dups = self.data_access.get_label_counts(workspace_id, dataset_name, category_id,
                                                                 remove_duplicates=True)
        self.assertEqual(labels_count_no_dups[str(LABEL_POSITIVE).lower()], len(sampled),
                         len(non_representative_duplicates))
        self.data_access.delete_all_labels(workspace_id, dataset_name)
        self.data_access.delete_dataset(dataset_name)


if __name__ == "__main__":
    unittest.main()



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import Mapping, Sequence

from label_sleuth.data_access.core.data_structs import TextElement


def convert_text_elements_to_train_data(elements: Sequence[TextElement], category_id) -> Sequence[Mapping]:
    """
    Convert a list of text elements to the expected format for training a model.
    :param elements: a list of TextElement objects
    :param category_id:
    :return: a list of dictionaries with at least the "text" and "label" fields, e.g. [{'text': 'text1', 'label': True,
    'additional_field': 'value1'}, {'text': 'text2', 'label': False,  'additional_field': 'value2'}]
    """
    labels = [element.category_to_label[category_id].label for element in elements]
    metadata = [element.category_to_label[category_id].metadata for element in elements]

    converted_data = [{"text": element.text, "label": label, **example_metadata}
                      for element, label, example_metadata in zip(elements, labels, metadata)]
    return converted_data





import os
import time
from collections import defaultdict

import pandas as pd

from typing import List
from label_sleuth.data_access.core.data_structs import Document, TextElement, DisplayFields, URI_SEP
from label_sleuth.data_access.processors.data_processor_api import DataProcessorAPI


def get_columns(df, column_names):
    column_values = []
    for col_name_to_add in column_names:
        if col_name_to_add is not None:
            if col_name_to_add not in df.columns:
                raise NameError(f'"{col_name_to_add}" is not one of the columns in the given DataFrame: {df.columns}')
            column_values.append(df[col_name_to_add].values)
        else:
            column_values.append([None]*len(df))
    return column_values


class CsvFileProcessor(DataProcessorAPI):
    """
    A DataProcessor for a corpus that is given in a csv format, one TextElement per line.

    """

    def __init__(self, dataset_name: str, temp_file_path: str, text_col: str = DisplayFields.text,
                 metadata_column_name_prefix: str = DisplayFields.csv_metadata_column_prefix,
                 doc_id_col: str = DisplayFields.doc_id, encoding: str = 'utf-8'):
        """

        :param dataset_name: the name of the processed dataset
        :param temp_file_path: the temporary file used to store the csv before processing
        :param text_col: the name of the column which holds the text of the TextElement. Default is 'text'.
        :param metadata_column_name_prefix: prefix of metadata columns that should be saved with the text element
        :param doc_id_col: column name by which text elements should be grouped into docs.
        If None all text elements would be put in a single dummy with the current time the format of '%d_%b_%Y_%H:%M'
         as the doc_id. Default is None.
        :param encoding: the encoding to use to read the csv raw file(s). Default is `utf-8`.

        """
        self.dataset_name = dataset_name
        self.temp_file_path = temp_file_path
        self.text_col = text_col
        self.metadata_column_name_prefix = metadata_column_name_prefix
        self.doc_id_col = doc_id_col
        self.encoding = encoding

    def get_raw_data_path(self) -> str:
        return self.temp_file_path

    def build_documents(self) -> List[Document]:
        """
        Process the raw data into a list of Documents.

        :rtype: List[Document]
        """
        if not os.path.isfile(self.get_raw_data_path()):
            raise Exception(f'file for dataset {self.dataset_name} not found')

        df = pd.read_csv(self.get_raw_data_path(), encoding=self.encoding, na_filter=False)
        df = df[df[self.text_col].apply(lambda x: len(x) > 0)]
        if self.doc_id_col not in df.columns:
            self.doc_id_col = None
        doc_uri_to_text_elements = defaultdict(list)
        metadata_column_names = [col for col in df.columns if col.startswith(self.metadata_column_name_prefix)]
        metadata_dict = {col_name.split(self.metadata_column_name_prefix)[1]: col_values
                         for col_name, col_values in zip(metadata_column_names, get_columns(df, metadata_column_names))}
        for idx, (text, doc_id) in enumerate(zip(*get_columns(df, [self.text_col, self.doc_id_col]))):
            if doc_id is None:
                doc_name_for_uri = time.strftime('%d_%b_%Y_%H-%M-%S', time.gmtime())
            else:
                doc_name_for_uri = str(doc_id).replace(URI_SEP, '_')
            doc_uri = self.dataset_name + URI_SEP + doc_name_for_uri
            if doc_uri not in doc_uri_to_text_elements:
                element_id = 0
                text_span_start = 0
            else:
                prev_element = doc_uri_to_text_elements[doc_uri][-1]
                element_id = int(prev_element.uri.split(URI_SEP)[-1]) + 1
                text_span_start = prev_element.span[0][1] + 1

            text_element_uri = doc_uri + URI_SEP + str(element_id)
            metadata = {k: v[idx] for k, v in metadata_dict.items()}
            text_element = TextElement(uri=text_element_uri, text=text,
                                       span=[(text_span_start, (text_span_start+len(text)))], metadata=metadata,
                                       category_to_label={})
            doc_uri_to_text_elements[doc_uri].append(text_element)

        documents = [Document(uri=doc_uri, text_elements=text_elements, metadata={})
                     for doc_uri, text_elements in doc_uri_to_text_elements.items()]
        return documents



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import os
import re
import shutil
import tempfile

from concurrent.futures import Future
from dataclasses import dataclass
from typing import Iterable, Mapping, Sequence, Tuple, List, Union

import numpy as np

from label_sleuth.models.core.model_api import ModelAPI
from label_sleuth.models.core.model_type import ModelType
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s')


@dataclass
class EnsemblePrediction(Prediction):
    model_type_to_prediction: dict


@dataclass
class EnsembleComponents:
    models: List


class Ensemble(ModelAPI):
    def __init__(self, output_dir, model_types: Iterable[ModelType],
                 background_jobs_manager: BackgroundJobsManager,
                 model_factory,
                 aggregation_func=lambda x: np.mean(x, axis=0)):
        """
        Create an ensemble model aggregating different model types

        :param output_dir:
        :param model_types: a list of model types
        :param model_factory:
        :param background_jobs_manager:
        :param aggregation_func: function to aggregate the predictions of the different models. The function gets an
        array of scores (where dimension 0 is the model types and dimension 1 is the list of elements), and returns a
        vector of aggregated scores. Defaults to mean score aggregation.
        """
        super().__init__(output_dir, background_jobs_manager)
        self.aggregation_func = aggregation_func
        self.model_types = model_types
        self.model_apis = [model_factory.get_model_api(model_type) for model_type in model_types]

    def train(self, train_data, language, model_params=None, done_callback=None) -> Tuple[str, Future]:
        """
        Submits training of the different model types in the background, and returns a model_id and future object
        for the ensemble model.
        """
        if model_params is None:
            model_params = {}
        model_ids_and_futures = [model_api.train(train_data, language, model_params) for model_api in self.model_apis]
        ensemble_model_id = ",".join(model_id for model_id, future in model_ids_and_futures)
        self.mark_train_as_started(ensemble_model_id)
        self.save_metadata(ensemble_model_id, language, model_params)

        future = self.background_jobs_manager.add_background_job(
            self.wait_and_update_status,
            args=(ensemble_model_id, [future for model_id, future in model_ids_and_futures]),
            use_gpu=self.gpu_support, done_callback=done_callback)
        logging.info(f"training an ensemble model id {ensemble_model_id} using {len(train_data)} elements")
        return ensemble_model_id, future

    def wait_and_update_status(self, model_id: str, train_futures: Iterable[Future]) -> str:
        """
        Wait for the training of the different models, and return *model_id* if all have finished successfully.

        :param model_id: id for the ensemble model
        :param train_futures: future objects for the _train() functions of the different models
        :return: model_id
        """

        try:
            for future in train_futures:
                future.result()
            self.mark_train_as_completed(model_id)
        except Exception:
            logging.exception(f'model {model_id} failed with exception')
            self.mark_train_as_error(model_id)
            raise

        return model_id

    def _train(self, model_id: str, train_data: Sequence[Mapping], model_params: dict):
        pass

    def load_model(self, model_path: Union[str, List]) -> EnsembleComponents:
        if isinstance(model_path, list):
            # for internal use by the system, where paths for the constituting models are passed
            # from Ensemble._infer_by_id()
            model_paths = model_path
        else:
            # for external use, where the ensemble has been exported into a single folder containing all the models
            model_path = model_path.rstrip("/")

            def replace_model_prefix(prefix):  # for backward compatibility of old model api class names
                return prefix.replace('_GloVe', '_WordEmbeddings')

            model_dir_names = [d for d in os.listdir(model_path) if os.path.isdir(os.path.join(model_path, d))]
            api_name_to_path = {}
            for model_dir_name in model_dir_names:
                dir_prefix = model_dir_name.rsplit('_', 1)[0]
                dir_prefix = replace_model_prefix(dir_prefix)
                api_name_to_path[dir_prefix] = os.path.join(model_path, model_dir_name)
            model_paths = [api_name_to_path[model_api.__class__.__name__] for model_api in self.model_apis]

        models = [model_api.load_model(model_path) for model_api, model_path in zip(self.model_apis, model_paths)]
        return EnsembleComponents(models=models)

    def _infer_by_id(self, model_id, items_to_infer):
        """
        We override ModelAPI._infer as we need to extract the paths to all of the models constituting the ensemble
        """
        model_paths = [model_api.get_model_dir_by_id(m_id)
                       for m_id, model_api in zip(model_id.split(","), self.model_apis)]
        model = self.load_model(model_path=model_paths)
        return self.infer(model, items_to_infer)

    def infer(self, ensemble: EnsembleComponents, items_to_infer) -> Sequence[EnsemblePrediction]:
        """
        Aggregate the predictions returned by the different models using self.aggregation_func
        """
        type_to_all_predictions = {}
        all_scores = []
        for model_api, model, model_type in zip(self.model_apis, ensemble.models, self.model_types):
            predictions = model_api.infer(model, items_to_infer)
            type_to_all_predictions[model_type.name] = predictions
            all_scores.append([pred.score for pred in predictions])
        aggregated_scores = np.apply_along_axis(self.aggregation_func, arr=np.array(all_scores), axis=0)
        labels = [score > 0.5 for score in aggregated_scores]
        type_to_prediction_per_element = [{model_type: model_preds[i] for model_type, model_preds
                                           in type_to_all_predictions.items()} for i in range(len(items_to_infer))]
        return [EnsemblePrediction(label=label, score=score, model_type_to_prediction=type_to_prediction)
                for label, score, type_to_prediction in zip(labels, aggregated_scores, type_to_prediction_per_element)]

    def delete_model(self, model_id):
        for model_api, m_id in zip(self.model_apis, model_id.split(",")):
            model_api.delete_model(m_id)

    def get_prediction_class(self):
        return EnsemblePrediction

    def copy_model_dir_for_export(self, ensemble_model_id):
        temp_path = str(tempfile.mkdtemp())
        ensemble_output_path = os.path.join(temp_path, os.path.basename(self.get_model_dir_by_id(ensemble_model_id)))
        for model_api, m_id in zip(self.model_apis, ensemble_model_id.split(",")):
            model_path = model_api.get_model_dir_by_id(m_id)
            output_path = os.path.join(ensemble_output_path, os.path.basename(model_path))
            shutil.copytree(model_path, output_path)
        return ensemble_output_path

    def get_supported_languages(self):
        return set.intersection(*[model_api.get_supported_languages() for model_api in self.model_apis])


class SVM_Ensemble(Ensemble):
    def __init__(self, output_dir, background_jobs_manager, model_factory):
        from label_sleuth.models.core.catalog import ModelsCatalog
        super().__init__(output_dir=output_dir, background_jobs_manager=background_jobs_manager,
                         model_types=[ModelsCatalog.SVM_OVER_BOW, ModelsCatalog.SVM_OVER_WORD_EMBEDDINGS],
                         model_factory=model_factory)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from dataclasses import dataclass


@dataclass
class Prediction:
    """
    An object containing model predictions for a single element.
    Each model.infer method should return at least the label and score fields. In order for a model to return
    additional fields (e.g model embeddings), a different dataclass that inherits from this one, and adds the desired
    fields, can be used.
    """
    label: bool
    score: float

    def __post_init__(self):
        # Since many models return numpy objects, which are not json-serializable, we convert them here
        self.label = bool(self.label)
        self.score = float(self.score)

        if self.score < 0 or self.score > 1:
            raise Exception(f'Model score {self.score} is outside the range [0-1]')



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import logging
import random
import sys

from collections import Counter
from typing import Sequence, List

from label_sleuth.data_access.core.data_structs import Label, LabelType, TextElement, BINARY_LABELS, LABEL_POSITIVE, \
    LABEL_NEGATIVE
from label_sleuth.training_set_selector.train_set_selector_api import TrainSetSelectorAPI


class TrainSetSelectorAllLabeled(TrainSetSelectorAPI):
    """
    Implements the basic model training behavior: use all elements labeled by the user - and only those elements -
    for training the model.
    """

    def __init__(self, data_access, background_jobs_manager, label_types=frozenset({LabelType.Standard})):
        super().__init__(data_access, background_jobs_manager, label_types=label_types)

    def get_train_set(self, workspace_id, train_dataset_name, category_id,
                      category_name: str, category_description: str) -> Sequence[TextElement]:
        train_data, train_counts = self.get_data_and_counts_for_labeled(workspace_id, train_dataset_name, category_id,
                                                                        remove_duplicates=True)
        self.verify_all_labels_are_in_train(train_counts)
        logging.info(f"using {len(train_data)} for train using dataset {train_dataset_name}")
        return train_data

    def get_data_and_counts_for_labeled(self, workspace_id, dataset_name, category_id, remove_duplicates=False):
        if dataset_name is None:
            return None, None
        labeled_elements = \
            self.data_access.get_labeled_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                                       category_id=category_id, sample_size=sys.maxsize,
                                                       remove_duplicates=remove_duplicates,
                                                       label_types=self.label_types)["results"]
        labels = [element.category_to_label[category_id].label for element in labeled_elements]
        counts = Counter(labels)

        return labeled_elements, counts

    @staticmethod
    def verify_all_labels_are_in_train(train_counts):
        labels_not_in_train = [label for label in BINARY_LABELS if train_counts[label] == 0]
        if len(labels_not_in_train) > 0:
            raise Exception(
                f"no train examples for labels: {labels_not_in_train}, cannot train a model: {train_counts}")


class TrainSetSelectorEnforcePositiveNegativeRatio(TrainSetSelectorAllLabeled):
    """
    Implements a training behavior that aims to place the ratio between positive and negative examples within a
    predetermined range.
    If the number of negative examples labeled by the user does not meet the minimal ratio, some _unlabeled_ examples
    are added to the train set as negative examples, and are marked as LabelType.Weak; If the number of negative
    examples labeled by the user exceeds the maximal ratio, only a sample of the user-labeled negative examples
    will be sent to the model.
    """
    def __init__(self, data_access, background_jobs_manager, label_types,
                 required_negative_ratio=None, max_negative_ratio=10**6):
        """
        :param data_access
        :param required_negative_ratio: required number of negative samples per positive
        :param max_negative_ratio: maximal allowed number of negative samples per positive
        """
        super().__init__(data_access, background_jobs_manager, label_types)
        self.negative_ratio = required_negative_ratio
        self.max_negative_ratio = max_negative_ratio
        self.neg_label = LABEL_NEGATIVE
        self.pos_label = LABEL_POSITIVE

    def get_train_set(self, workspace_id, train_dataset_name, category_id,
                      category_name: str, category_description: str) -> Sequence[TextElement]:
        train_data, train_counts = self.get_data_and_counts_for_labeled(workspace_id, train_dataset_name, category_id,
                                                                        remove_duplicates=True)
        current_neg_count = train_counts.get(self.neg_label, 0)
        required_neg_count = self.negative_ratio * train_counts[self.pos_label]
        max_neg_count = self.max_negative_ratio * train_counts[self.pos_label]

        if current_neg_count < required_neg_count:
            required_number_of_unlabeled_as_neg = required_neg_count - current_neg_count
            logging.info(f"Trying to add {required_number_of_unlabeled_as_neg} to meet ratio of "
                         f"{self.negative_ratio} negatives per positive")
            # Weak negatives are added according to a fixed order for the specific dataset
            elements_by_order = self.get_weak_negative_candidates(workspace_id, train_dataset_name, category_id)
            weak_negatives_added = 0
            for element in elements_by_order:
                if weak_negatives_added == required_number_of_unlabeled_as_neg:
                    break
                if category_id not in element.category_to_label:  # unlabeled
                    element.category_to_label = {category_id: Label(self.neg_label, label_type=LabelType.Weak)}
                    train_data.append(element)
                    weak_negatives_added += 1
            train_counts[self.neg_label] += weak_negatives_added

        elif current_neg_count > max_neg_count:
            logging.info(f"Too many strong negative elements, sampling to get to a ratio of "
                         f"less than {self.max_negative_ratio} negatives per positive")
            positives = [element for element in train_data
                         if element.category_to_label[category_id].label == self.pos_label]
            shuffled_elements = get_elements_by_selection_order(workspace_id, train_dataset_name, self.data_access)
            negatives_to_keep = [element for element in shuffled_elements if category_id in element.category_to_label
                                 and element.category_to_label[category_id].label == self.neg_label][:max_neg_count]
            train_data = positives + negatives_to_keep
            train_counts[self.neg_label] = len(negatives_to_keep)

        self.verify_all_labels_are_in_train(train_counts)

        logging.info(f"using {len(train_data)} for train using dataset {train_dataset_name}")
        return train_data

    def get_weak_negative_candidates(self, workspace_id, dataset_name, category_id) -> List[TextElement]:
        return get_elements_by_selection_order(workspace_id, dataset_name, self.data_access)


def get_elements_by_selection_order(workspace_id, dataset_name, data_access):
    dataset_uris = sorted(data_access.get_all_text_elements_uris(dataset_name))
    random_seed = sum([ord(c) for c in dataset_name])
    random.Random(random_seed).shuffle(dataset_uris)
    elements_by_order = data_access.get_text_elements_by_uris(workspace_id=workspace_id, dataset_name=dataset_name,
                                                              uris=dataset_uris)
    return elements_by_order



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc

from typing import Sequence, Type

import numpy as np

from label_sleuth.data_access.core.data_structs import TextElement
from label_sleuth.models.core.prediction import Prediction


class ActiveLearner:
    """
    Base class for implementing an active learning module.
    """
    def get_recommended_items_for_labeling(self, workspace_id: str, dataset_name: str, category_id: int,
                                           candidate_text_elements: Sequence[TextElement],
                                           candidate_text_element_predictions: Sequence[Prediction],
                                           sample_size: int = 1) -> Sequence[TextElement]:
        """
        Returns a batch of *sample_size* elements suggested by the active learning module,
        for a given dataset and category, based on the model predictions for a sequence of candidate text elements.
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        :param candidate_text_elements:
        :param candidate_text_element_predictions:
        :param sample_size: number of suggested elements to return
        """
        scores = self.get_per_element_score(candidate_text_elements, candidate_text_element_predictions,
                                            workspace_id, dataset_name, category_id)
        sorted_indices = np.argsort(scores)[::-1]
        top_indices = sorted_indices[:sample_size]
        recommended_items = np.array(candidate_text_elements)[np.array(top_indices)].tolist()
        return recommended_items

    @abc.abstractmethod
    def get_per_element_score(self, candidate_text_elements: Sequence[TextElement],
                              candidate_text_element_predictions: Sequence[Prediction],
                              workspace_id: str, dataset_name: str, category_id: int) -> Sequence[float]:
        """
        For a given sequence of TextElements and their model predictions, return a score for each element
        from the active learning module.
        :param candidate_text_elements:
        :param candidate_text_element_predictions:
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        """


class ActiveLearningStrategy:
    def __init__(self, cls: Type[ActiveLearner]):
        self.cls = cls
        self.name = cls.__name__



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import Sequence

from label_sleuth.active_learning.core.active_learning_api import ActiveLearner
from label_sleuth.data_access.core.data_structs import TextElement
from label_sleuth.models.core.prediction import Prediction


class HardMiningLearner(ActiveLearner):
    """
    This active learning module suggests "hard" examples, i.e. examples that the model is most uncertain about.
    The highest active learning scores are given to examples which were given model scores closest to the
    decision threshold of 0.5.
    """
    def get_per_element_score(self, candidate_text_elements: Sequence[TextElement],
                              candidate_text_element_predictions: Sequence[Prediction], workspace_id: str,
                              dataset_name: str, category_id: int) -> Sequence[float]:
        return [2*(0.5-abs(pred.score-0.5)) for pred in candidate_text_element_predictions]



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
import random
import numpy as np

from label_sleuth.models.core.languages import Languages
from label_sleuth.models.core.model_api import ModelAPI, ModelStatus
from label_sleuth.models.core.prediction import Prediction
from label_sleuth.orchestrator.background_jobs_manager import BackgroundJobsManager


class RandomModel(ModelAPI):
    """
    Mock classification model that does not train, and returns random classification predictions.
    """
    def __init__(self, output_dir, background_jobs_manager: BackgroundJobsManager):
        super().__init__(output_dir, background_jobs_manager)

        self.model_id_to_random_seed = {}
        self.random_seed = -1

    def _train(self, model_id, train_data, model_params):
        seed = self.random_seed + 1
        self.model_id_to_random_seed[model_id] = seed
        self.random_seed = seed

    def load_model(self, model_path) -> int:
        # RandomModel does not require loading any model objects from disk, we implement it this way for compatibility
        # with the rest of the system
        model_id = os.path.basename(model_path)
        random_seed = self.model_id_to_random_seed[model_id]
        return random_seed

    def infer(self, random_seed: int, items_to_infer):
        rand = random.Random(random_seed)
        scores = np.array([rand.random() for _ in range(len(items_to_infer))])
        labels = [score > 0.5 for score in scores]
        return [Prediction(label=label, score=score) for label, score in zip(labels, scores)]

    def get_model_status(self, model_id):
        if model_id in self.model_id_to_random_seed:
            return ModelStatus.READY
        return ModelStatus.ERROR

    def delete_model(self, model_id):
        if model_id in self.model_id_to_random_seed:
            self.model_id_to_random_seed.pop(model_id)

    def get_supported_languages(self):
        return Languages.all_languages()



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import ast
import dataclasses
import logging
import os
import random
import shutil
import sys
import threading
import time

import jsonpickle
import ujson as json
import pandas as pd

from pathlib import Path
from collections import Counter, defaultdict
from typing import Sequence, Iterable, Mapping, List, Union, Set

import label_sleuth.data_access.file_based.utils as utils
from label_sleuth.data_access.core.data_structs import Document, Label, TextElement, LabelType
from label_sleuth.data_access.data_access_api import DataAccessApi, AlreadyExistsException, DocumentStatistics, \
    LabeledStatus


class FileBasedDataAccess(DataAccessApi):
    """
    Under the FileBasedDataAccess implementation of DataAccessApi, information about labels and TextElements is stored
    in the file system, as well as in memory, in the variables "ds_in_memory" and "labels_in_memory" below.
    Note that while label information changes over time (as labels are added, altered etc.), the TextElement information
    inside "ds_in_memory" only changes if new documents are added to the dataset.

    ===ds_in_memory===
    maps dataset_name to a pandas DataFrame containing all the dataset text elements

    ===labels_in_memory===
    maps workspace_id -> dataset name -> URIs -> categories -> Label object

    """
    doc_dir_name = 'doc_dump'
    sentences_filename = 'dataset_sentences.csv'
    labels_filename = 'workspace_labels.json'

    workspace_to_labels_lock_objects = defaultdict(threading.Lock)
    ds_in_memory = defaultdict(pd.DataFrame)
    labels_in_memory = defaultdict(lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(Label))))
    dataset_in_memory_lock = threading.RLock()

    def __init__(self, output_dir):
        self.output_dir = output_dir

    def add_documents(self, dataset_name: str, documents: Iterable[Document]):
        """
        Add new documents to a given dataset; If dataset does not exist, create it.

        In this implementation, the same data is stored in two formats:
            A. a json file per Document, containing data for the TextElement objects within that Document
            B. a single csv file, containing data for the TextElement objects from all Documents for this dataset

        :param dataset_name: the name of the dataset to which the documents should be added.
        :param documents: an Iterable over Document type.
        """
        doc_dump_dir = self._get_documents_dump_dir(dataset_name)
        sentences = []
        if not os.path.exists(doc_dump_dir):
            os.makedirs(doc_dump_dir)
        else:
            doc_ids = {document.uri for document in documents}
            intersection = doc_ids.intersection(set(self.get_all_document_uris(dataset_name)))
            if len(intersection) > 0:
                raise AlreadyExistsException(f"{len(intersection)} documents are already in dataset '{dataset_name}'. "
                                             f"uris: ({list(intersection)[:5]}{'...' if len(intersection) > 5 else ''})",
                                             list(intersection))

        for doc in documents:
            # save doc to file
            filename = utils.uri_to_filename(doc.uri)
            file_path = os.path.join(doc_dump_dir, filename)
            doc_encoded = jsonpickle.encode(doc)
            with open(file_path + '.json', 'w') as f:
                f.write(doc_encoded)
            # add doc sentences to working memory
            sentences.extend(doc.text_elements)
        self._add_sentences_to_dataset_in_memory(dataset_name=dataset_name, text_elements=sentences)

        num_of_text_elements = sum([len(doc.text_elements) for doc in documents])
        logging.info(f'{dataset_name}:\t\tloaded {len(documents)} documents '
                     f'({num_of_text_elements} text elements) under {doc_dump_dir}')
        return DocumentStatistics(len(documents), num_of_text_elements)

    def set_labels(self, workspace_id: str, uris_to_labels: Mapping[str, Mapping[int, Label]],
                   apply_to_duplicate_texts=False):
        """
        Set labels to TextElements in dataset for a given workspace_id.

        :param workspace_id: the workspace_id of the labeling effort.
        :param uris_to_labels: list of tuples of TextElement URI and a dict that represents a label.
        The dict keys are category ids and values are Labels. For example: [(uri_1, {0: Label_cat_1}),
                                                                            (uri_2, {0: Label_cat_1,
                                                                                     1: Label_cat_2})]
        :param apply_to_duplicate_texts: if True, also set the same labels for additional URIs that are duplicates
        of the URIs provided.
        """

        # Assuming all elements are from the same dataset, using the first element's dataset name
        dataset_name = utils.get_dataset_name_from_uri(next(iter(uris_to_labels.keys())))
        with self._get_lock_object_for_workspace(workspace_id):
            ds_labels = self._get_labels(workspace_id, dataset_name)
            all_uris = self.get_all_text_elements_uris(dataset_name)
            for uri, labels in uris_to_labels.items():
                if uri not in all_uris:
                    raise Exception(f'Trying to set labels for uri "{uri}" which does not exist')

                if apply_to_duplicate_texts:  # set the given label for all elements with the same text
                    same_text_uris = self._get_uris_with_the_same_text(dataset_name, uri)
                    for same_text_uri in same_text_uris:
                        ds_labels[same_text_uri].update(labels)
                else:
                    # Note: we do not override existing labels if they are from another category
                    ds_labels[uri].update(labels)
            # Save updated labels dict to disk
            self._save_labels_data(dataset_name, workspace_id)

    def unset_labels(self, workspace_id: str, category_id: int, uris: Sequence[str], apply_to_duplicate_texts=False):
        """
        Remove workspace labels for a certain category from a specified list of uris.

        :param workspace_id: the workspace_id of the labeling effort.
        :param category_id: the id of the category labels are assigned to.
        :param uris: list of URIs to unset the label for.
        :param apply_to_duplicate_texts: if True, also unset the same labels for additional URIs that have the same text
        as that of the URIs provided.
        """
        dataset_name = utils.get_dataset_name_from_uri(uris[0])
        with self._get_lock_object_for_workspace(workspace_id):
            ds_labels = self._get_labels(workspace_id, dataset_name)
            all_uris = self.get_all_text_elements_uris(dataset_name)
            for uri in uris:
                if uri not in all_uris:
                    raise Exception(f'Trying to unset labels for uri "{uri}" which does not exist')

                if apply_to_duplicate_texts:  # unset the given label for all elements with the same text
                    same_text_uris = self._get_uris_with_the_same_text(dataset_name, uri)
                    for same_text_uri in same_text_uris:
                        ds_labels[same_text_uri].pop(category_id)
                        if len(ds_labels[same_text_uri]) == 0:
                            ds_labels.pop(same_text_uri)
                else:
                    ds_labels[uri].pop(category_id)
                    if len(ds_labels[uri]) == 0:
                        ds_labels.pop(uri)
            # Save updated labels dict to disk
            self._save_labels_data(dataset_name, workspace_id)

    def get_documents(self, workspace_id: Union[None, str], dataset_name: str, uris: Iterable[str],
                      label_types: Union[None,Set[LabelType]] = frozenset({LabelType.Standard})) \
            -> List[Document]:
        """
        Return a List of Documents, from the given dataset_name, matching the uris provided, and add the label
        information for the workspace to the TextElements of these Documents, if available.

        :param workspace_id: the workspace_id of the labeling effort. if None, documents are returned without
        label information
        :param dataset_name: the name of the dataset from which the documents should be retrieved.
        :param uris: an Iterable of uris of Documents, represented as string.
        :param label_types:  If workspace_id is provided, select which label_types to retrieve
                     by default, only the LabelType.Standard (strong labels) are retrieved.
        :return: a List of Document objects, from the given dataset_name, matching the uris provided, containing label
        information for the TextElements of these Documents, if available.
        """

        def load_doc(uri):
            filename = utils.uri_to_filename(uri)
            file_path = os.path.join(doc_dump_dir, filename)
            with open(file_path + '.json') as json_file:
                doc_encoded = json_file.read()
            doc = jsonpickle.decode(doc_encoded)
            return doc

        doc_dump_dir = self._get_documents_dump_dir(dataset_name)
        docs = [load_doc(uri) for uri in uris]
        if workspace_id is not None:
            with self._get_lock_object_for_workspace(workspace_id):
                for d in docs:
                    self._add_labels_info_for_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                                            text_elements=d.text_elements, label_types=label_types)
        return docs

    def get_all_document_uris(self, dataset_name: str) -> List[str]:
        """
        Return a List of all Document uris in the given dataset_name.

        :param dataset_name: the name of the dataset from which the Document uris should be retrieved.
        :return: a List of all Document uris in the given dataset_name.
        """
        uris = []
        doc_dump_dir = self._get_documents_dump_dir(dataset_name)
        for filename in os.listdir(doc_dump_dir):
            filename, file_extension = os.path.splitext(filename)
            uris.append(utils.filename_to_uri(filename))
        uris = sorted(uris, key=utils.get_sort_key_by_document_name)
        return uris

    def get_all_text_elements_uris(self, dataset_name: str) -> List[str]:
        """
        Return a List of all TextElement uris in the given dataset_name.

        :param dataset_name: the name of the dataset from which the TextElement uris should be retrieved.
        :return: a List of all TextElement uris in the given dataset_name.
        """
        return list(self._get_ds_in_memory(dataset_name)['uri'].values)

    def get_text_element_count(self, dataset_name: str) -> int:
        """
        Return the total number of TextElements in the given dataset_name.
        :param dataset_name:
        """
        return len(self.get_all_text_elements_uris(dataset_name))

    def get_all_text_elements(self, dataset_name: str) -> List[TextElement]:
        """
        Return a List of all TextElement in the given dataset_name.

        :param dataset_name: the name of the dataset from which the TextElement should be retrieved.
        """
        return utils.build_text_elements_from_dataframe_and_labels(self._get_ds_in_memory(dataset_name), labels_dict={})

    def get_text_elements(self, workspace_id: str, dataset_name: str, sample_size: int = sys.maxsize,
                          sample_start_idx: int = 0, query: str = None, is_regex: bool = False, document_uri=None,
                          remove_duplicates=False, random_state: int = 0) -> Mapping:
        """
        Sample *sample_size* TextElements from dataset_name, optionally limiting to those matching a query,
        and add their labels information for workspace_id, if available.

        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which TextElements are sampled
        :param sample_size: how many TextElements should be sampled
        :param sample_start_idx: get elements starting from this index (for pagination). Default is 0
        :param query: a query string to search for in the sampled TextElements.
                      If None, then no such filtering is performed.
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param document_uri: get elements from a particular document
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """
        with self._get_lock_object_for_workspace(workspace_id):
            results_dict = \
                self._get_text_elements(
                    workspace_id=workspace_id, dataset_name=dataset_name,
                    filter_func=lambda df, _: utils.filter_by_query_and_document_uri(df, query, is_regex, document_uri),
                    sample_size=sample_size, sample_start_idx=sample_start_idx,
                    remove_duplicates=remove_duplicates, random_state=random_state)

        return results_dict

    def get_unlabeled_text_elements(self, workspace_id: str, dataset_name: str, category_id: int,
                                    sample_size: int = sys.maxsize, sample_start_idx: int = 0,
                                    query: str = None, is_regex: bool = False,
                                    remove_duplicates=False, random_state: int = 0) -> Mapping:
        """
        Sample *sample_size* TextElements from dataset_name, unlabeled for category_id in workspace_id, optionally
        limiting to those matching a query.

        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which TextElements are sampled
        :param category_id: we demand that the elements are not labeled for this category
        :param sample_size: how many TextElements should be sampled
        :param sample_start_idx: get elements starting from this index (for pagination). Default is 0
        :param query: a query string to search for in the sampled TextElements.
                      If None, then no such filtering is performed.
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """
        filter_func = lambda df, labels: \
            utils.filter_by_query_and_label_status(df, labels, category_id, LabeledStatus.UNLABELED, query, is_regex)

        with self._get_lock_object_for_workspace(workspace_id):
            results_dict = self._get_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                                   filter_func=filter_func, sample_size=sample_size,
                                                   sample_start_idx=sample_start_idx,
                                                   remove_duplicates=remove_duplicates, random_state=random_state)
        return results_dict

    def get_labeled_text_elements(self, workspace_id: str, dataset_name: str, category_id: int,
                                  sample_size: int = sys.maxsize, query: str = None, is_regex: bool = False,
                                  remove_duplicates=False, random_state: int = 0,
                                  label_types: Set[LabelType] = frozenset({LabelType.Standard})) -> Mapping:
        """
        Sample *sample_size* TextElements from dataset_name, labeled for category_id in workspace_id,
        optionally limiting to those matching a query.

        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which TextElements are sampled
        :param category_id: we demand that the elements are labeled for this category
        :param sample_size: how many TextElements should be sampled
        :param query: a query string to search for in the sampled TextElements.
                      If None, then no such filtering is performed.
        :param is_regex: if True, the query string is interpreted as a regular expression (False by default)
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :param label_types: by default, only the LabelType.Standard (strong labels) are retrieved.
        :return: a dictionary with two keys: 'results' whose value is a list of TextElements, and 'hit_count' whose
        value is the total number of TextElements in the dataset matched by the query.
        {'results': [TextElement], 'hit_count': int}
        """

        filter_func = lambda df, labels: \
            utils.filter_by_query_and_label_status(df, labels, category_id, LabeledStatus.LABELED, query, is_regex,
                                                   label_types=label_types)
        with self._get_lock_object_for_workspace(workspace_id):
            results_dict = self._get_text_elements(workspace_id=workspace_id, dataset_name=dataset_name,
                                                   filter_func=filter_func, sample_size=sample_size,
                                                   remove_duplicates=remove_duplicates, random_state=random_state)
        return results_dict

    def get_label_counts(self, workspace_id: str, dataset_name: str, category_id: int, remove_duplicates=False,
                         label_types: Set[LabelType] = frozenset(LabelType._member_map_.values()),
                         fine_grained_counts=True) -> Mapping[Union[str, bool], int]:
        """
        Return for each label value, assigned to category_id, the total count of its appearances in dataset_name.
        :param workspace_id: the workspace_id of the labeling effort.
        :param dataset_name: the name of the dataset from which labels count should be generated
        :param category_id: the id of the category whose label information is the target
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        :param label_types: by default, labels of all types are retrieved.
        :param fine_grained_counts: if True, count labels of each label type separately.
        :return: a map whose keys are label values, and the values are the number of TextElements this label was
        assigned to.
        """
        labels_by_uri = self._get_labels(workspace_id=workspace_id, dataset_name=dataset_name).copy()
        if remove_duplicates:
            corpus_df = self._get_ds_in_memory(dataset_name=dataset_name)
            labeled_df = corpus_df[corpus_df['uri'].isin(labels_by_uri)]
            uris_to_keep = set(labeled_df.drop_duplicates(subset=['text'])['uri'])
            labels_by_uri = {uri: category_to_label for uri, category_to_label in labels_by_uri.items()
                             if uri in uris_to_keep}

        category_label_list = \
            [category_to_label[category_id] for category_to_label in labels_by_uri.values()
             if category_id in category_to_label and category_to_label[category_id].label_type in label_types]

        if fine_grained_counts:
            return Counter(lbl_obj.get_detailed_label_name() for lbl_obj in category_label_list)
        else:
            return Counter(lbl_obj.label for lbl_obj in category_label_list)

    def delete_all_labels(self, workspace_id, dataset_name):
        """
        Delete the labels info of the given workspace_id for the given dataset (other labels info files are kept).
        :param workspace_id:
        :param dataset_name:
        """
        labels_file = self._get_workspace_labels_dump_filename(workspace_id, dataset_name)
        if os.path.isfile(labels_file):
            os.remove(labels_file)
        workspace_dumps_dir = self._get_workspace_labels_dir(workspace_id)
        if os.path.exists(workspace_dumps_dir) and len(os.listdir(workspace_dumps_dir)) == 0:
            os.rmdir(workspace_dumps_dir)

    def delete_labels_for_category(self, workspace_id, dataset_name, category_id):
        """
        Delete the labels info associated with the given category.
        :param workspace_id:
        :param dataset_name:
        :param category_id:
        """
        labeled_elements = self.get_labeled_text_elements(workspace_id, dataset_name, category_id,
                                                          sample_size=sys.maxsize)['results']
        if len(labeled_elements) > 0:
            self.unset_labels(workspace_id, category_id, [e.uri for e in labeled_elements])

    def get_text_elements_by_uris(self, workspace_id: str, dataset_name: str, uris: Iterable[str],
                                  label_types: Set[LabelType] = frozenset({LabelType.Standard}))\
                                                                                    -> List[TextElement]:
        """
        Return a List of TextElement objects from the given dataset_name, matching the uris provided, and add the label
        information for the workspace to these TextElements, if available.
        :param workspace_id:
        :param dataset_name:
        :param uris:
        :param label_types:  by default, only the LabelType.Standard (strong labels) are retrieved.
        """
        corpus_df = self._get_ds_in_memory(dataset_name)
        uris = list(uris)
        corpus_df = corpus_df.loc[corpus_df['uri'].isin(uris)]
        text_elements_by_uri = {te.uri: te for te
                                in utils.build_text_elements_from_dataframe_and_labels(corpus_df, labels_dict={})}
        text_elements = [text_elements_by_uri.get(uri) for uri in uris]

        with self._get_lock_object_for_workspace(workspace_id):
            text_elements = self._add_labels_info_for_text_elements(workspace_id,
                                                                    dataset_name, text_elements, label_types)

        return text_elements

    def get_text_element_iterator(self, workspace_id, dataset_name, shuffle=False, random_state: int = 0,
                                  remove_duplicates=False) -> Iterable[TextElement]:
        """
        Enables iterating over TextElement objects from the given *dataset_name*. Particularly useful where the total
        number of elements required is not known in advance.
        :param workspace_id:
        :param dataset_name:
        :param shuffle: if True, the iterator yields the text elements in random order.
        :param random_state: provide an int seed to define a random state. Default is zero.
        :param remove_duplicates: if True, do not include elements that are duplicates of each other.
        """
        corpus_df = self._get_ds_in_memory(dataset_name)
        if remove_duplicates:
            corpus_df = corpus_df.drop_duplicates(subset=['text'])
        all_uris = list(corpus_df['uri'].values)
        if shuffle:
            random.Random(random_state).shuffle(all_uris)

        # extracting one element at a time from a large dataframe can be expensive, so we fetch them in batches
        # but yield them one by one
        batch_size = 1000
        for i in range(0, len(all_uris), batch_size):
            batch_uris = all_uris[i:i + batch_size]
            yield from self.get_text_elements_by_uris(workspace_id, dataset_name, batch_uris)

    def get_all_dataset_names(self) -> List[str]:
        """
        :return: a list of all available dataset names
        """
        path = self._get_datasets_base_dir()
        return [name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name))]

    def delete_dataset(self, dataset_name: str):
        """
        Delete dataset by name
        :param dataset_name:
        """
        logging.info(f"Deleting dataset '{dataset_name}'")
        dataset_dir = self._get_dataset_base_dir(dataset_name)
        if os.path.isdir(dataset_dir):
            shutil.rmtree(dataset_dir)
        if dataset_name in self.ds_in_memory:
            del self.ds_in_memory[dataset_name]

    def _get_lock_object_for_workspace(self, workspace_id: str):
        lock_object = self.workspace_to_labels_lock_objects[workspace_id]
        return lock_object

    def _get_ds_in_memory(self, dataset_name):
        with self.dataset_in_memory_lock:
            if dataset_name not in self.ds_in_memory:
                if self._dataset_exists(dataset_name):
                    logging.info(f"reading dataset '{dataset_name}' csv file")
                    dataset_file_path = self._get_dataset_dump_filename(dataset_name)

                    df = pd.read_csv(dataset_file_path, converters=
                            {"span": lambda span: [tuple(int(x) for x in span[2:-2].split(','))],
                             "metadata": lambda metadata: ast.literal_eval(metadata) if metadata != '{}' else {}})

                    logging.info(f"csv file for dataset '{dataset_name}' read and processed successfully")
                else:
                    raise Exception(f'Dataset "{dataset_name}" does not exist.')
                self.ds_in_memory[dataset_name] = df
            res = self.ds_in_memory[dataset_name]
        return res

    def _get_labels(self, workspace_id, dataset_name):
        if workspace_id not in self.labels_in_memory or dataset_name not in self.labels_in_memory[workspace_id]:
            file_path = self._get_workspace_labels_dump_filename(workspace_id, dataset_name)
            if os.path.isfile(file_path):
                # Read dict from disk
                with open(file_path) as f:
                    labels_encoded = f.read()
                simplified_dict = json.loads(labels_encoded)
                for uri, category_to_label in simplified_dict.items():
                    for category_id, label_dict in category_to_label.items():
                        self.labels_in_memory[workspace_id][dataset_name][uri][int(category_id)] = Label(**label_dict)
            else:
                # Save empty dict to disk
                os.makedirs(Path(file_path).parent, exist_ok=True)
                empty_dict_encoded = json.dumps(self.labels_in_memory[workspace_id][dataset_name])
                with open(file_path, 'w') as f:
                    f.write(empty_dict_encoded)
        return self.labels_in_memory[workspace_id][dataset_name]

    def _add_sentences_to_dataset_in_memory(self, dataset_name, text_elements: Iterable[TextElement]):
        if self._dataset_exists(dataset_name):
            self.ds_in_memory[dataset_name] = self._get_ds_in_memory(dataset_name)
        dicts = [dataclasses.asdict(s) for s in text_elements]
        [d.pop('category_to_label') for d in dicts]  # category_to_labels is not saved in the dataframe
        new_sentences_df = pd.DataFrame(dicts)
        if dataset_name in self.ds_in_memory:
            self.ds_in_memory[dataset_name] = self.ds_in_memory[dataset_name].append(new_sentences_df, sort=False)
        else:
            self.ds_in_memory[dataset_name] = new_sentences_df
        self.ds_in_memory[dataset_name] = self._add_text_unique_ids(self.ds_in_memory[dataset_name])
        self.ds_in_memory[dataset_name].to_csv(self._get_dataset_dump_filename(dataset_name), index=False)

    def _add_labels_info_for_text_elements(self, workspace_id, dataset_name, text_elements: List[TextElement],
                                           label_types):
        labels_info_for_workspace = self._get_labels(workspace_id, dataset_name)
        for elem in text_elements:
            if elem.uri in labels_info_for_workspace:
                if label_types is None:
                    elem.category_to_label = labels_info_for_workspace[elem.uri].copy()
                else:
                    elem.category_to_label = {cat:label for cat, label in labels_info_for_workspace[elem.uri].items()
                                              if label.label_type in label_types}
        return text_elements

    def _get_text_elements(self, workspace_id: str, dataset_name: str, filter_func, sample_size: int,
                           sample_start_idx=0, remove_duplicates=False, random_state: int = 0) -> Mapping:
        """
        :param workspace_id: if None no labels info would be used or output
        :param dataset_name:
        :param filter_func:
        :param sample_size: number of elements to return. if None, return all elements without sampling
        :param sample_start_idx: get elements starting from this index (for pagination)
        :param remove_duplicates:
        :param random_state: provide an int seed to define a random state. Default is zero.
        """
        corpus_df = self._get_ds_in_memory(dataset_name)
        if workspace_id:
            labels_dict = self._get_labels(workspace_id, dataset_name)
        else:
            labels_dict = {}
        labels_series = corpus_df['uri'].apply(lambda u: labels_dict[u] if u in labels_dict else {})
        corpus_df = filter_func(corpus_df, labels_series)

        results_dict = {'hit_count': len(corpus_df)}
        if remove_duplicates:
            unique_texts = set()
            corpus_df = corpus_df.loc[
                corpus_df['text'].apply(lambda x: x not in unique_texts and not unique_texts.add(x))]
            results_dict['hit_count_unique'] = len(corpus_df)

        if sample_size is not None:
            # TODO UNKNOWN BUG fix
            corpus_df = corpus_df.sample(n=min(sample_size + sample_start_idx, len(corpus_df)),
                                         random_state=random_state
                                         )[sample_start_idx:sample_start_idx + sample_size]

        results_dict['results'] = utils.build_text_elements_from_dataframe_and_labels(corpus_df, labels_dict)
        return results_dict

    def _save_labels_data(self, dataset_name, workspace_id):
        file_path = self._get_workspace_labels_dump_filename(workspace_id, dataset_name)
        os.makedirs(Path(file_path).parent, exist_ok=True)
        labels = self.labels_in_memory[workspace_id][dataset_name]
        simplified_labels = {k: {str(category_id): label.to_dict() for category_id, label in v.items()}
                             for k, v in labels.items()}
        labels_in_memory_encoded = json.dumps(simplified_labels)
        with open(file_path, 'w') as f:
            f.write(labels_in_memory_encoded)

    @staticmethod
    def _add_text_unique_ids(df):
        """
        To facilitate extraction of duplicate elements, i.e. text elements that have the same text, we assign an id to
        each unique text in the dataframe. Whenever new documents are added to the corpus, these ids are recalculated.
        """
        unique_to_id = {}
        df['text_unique_id'] = df['text'].apply(
            lambda x: len(unique_to_id) - 1 if x not in unique_to_id and not unique_to_id.update({x: len(unique_to_id)})
            else unique_to_id[x])
        return df

    def _get_uris_with_the_same_text(self, dataset_name, uri):
        ds_in_memory = self._get_ds_in_memory(dataset_name)
        text_unique_id = ds_in_memory[ds_in_memory['uri'] == uri]['text_unique_id'].values[0]
        return ds_in_memory[ds_in_memory['text_unique_id'] == text_unique_id]['uri']

    def _dataset_exists(self, dataset_name):
        return os.path.exists(self._get_dataset_dump_filename(dataset_name))

    def _get_dataset_base_dir(self, dataset_name):
        return os.path.join(self._get_datasets_base_dir(), dataset_name)

    def _get_datasets_base_dir(self):
        data_access_dumps_path = os.path.join(self.output_dir, 'data_access_dumps')
        os.makedirs(data_access_dumps_path, exist_ok=True)
        return data_access_dumps_path

    def _get_documents_dump_dir(self, dataset_name):
        return os.path.join(self._get_dataset_base_dir(dataset_name), self.doc_dir_name)

    def _get_dataset_dump_filename(self, dataset_name):
        return os.path.join(self._get_dataset_base_dir(dataset_name), self.sentences_filename)

    def _get_workspace_labels_dump_filename(self, workspace_id, dataset_name):
        workspace_dir = self._get_workspace_labels_dir(workspace_id)
        return os.path.join(workspace_dir, str(dataset_name) + '_' + self.labels_filename)

    def _get_workspace_labels_dir(self, workspace_id):
        return os.path.join(self.output_dir, 'user_labels', str(workspace_id))

    def preload_dataset(self, dataset_name):
        self._get_ds_in_memory(dataset_name)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import setuptools
import os


def package_files(prefix, directory):
    paths = []
    for (path, directories, filenames) in os.walk(os.path.join(prefix, directory)):
        for filename in filenames:
            paths.append(os.path.join(path, filename)[len(prefix)+1:])
    return paths


build_files = package_files("label_sleuth", "build")

requirements_file = 'requirements.txt'

# read requirements from file
with open(requirements_file) as fh:
    requirements = fh.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="label-sleuth",
    version=f"v0.9.4".replace('v', ''),
    author="IBM Research",
    author_email="eyals@il.ibm.com",
    url="https://github.com/label-sleuth/label-sleuth",
    description="Label Sleuth",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    packages=setuptools.find_packages(),
    license='Apache License 2.0',
    python_requires='>=3.8',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
    ],
    package_data={"": ["LICENSE", "config.json", "requirements.txt"] + build_files},
    include_package_data=True
)



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import abc
from typing import Set

from label_sleuth.models.core.model_type import ModelType


class ModelPolicy(object, metaclass=abc.ABCMeta):
    """
    Base class for implementing a model policy, that determines which type of classification model will be used.
    Policies can be static, i.e. always return the same model type, or dynamic, i.e. a different model type is returned
    depending on the current iteration.
    """

    @abc.abstractmethod
    def get_model_type(self, iteration_num: int) -> ModelType:
        """
        Given *iteration_num*, return the type of classification model to be used
        :param iteration_num:
        :return: a member of the ModelTypes enum
        """

    @abc.abstractmethod
    def get_all_model_types(self) -> Set[ModelType]:
        """
        Return the types of all possible classification models to be used by this policy. Useful for checking that
        the policy is compatible with the overall system configuration.
        """

    @abc.abstractmethod
    def get_name(self) -> str:
        """
        :return: a name that describes the policy
        """



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from typing import List

import numpy as np

from label_sleuth.models.core.model_type import ModelType
from label_sleuth.models.policy.model_policy import ModelPolicy


class ModelChangingPolicy(ModelPolicy):
    """
    A dynamic model policy that follows a predefined pattern of switching between model types after a certain number of
    iterations, e.g. for the first 3 iterations use ModelTypes.NB_OVER_BOW, then switch to using ModelTypes.HF_BERT
    """

    def __init__(self, model_types: List[ModelType], num_iterations_per_model: List[int]):
        """
        :param model_types: a list of the N model types to be used by the policy
        :param num_iterations_per_model: a corresponding list specifying the number of iterations for each of the
        first N-1 models to be used; the Nth model type will be used for all subsequent iterations.
        For example, if there are 3 model types, and num_iterations_per_model=[5, 2], then type A will be used for
        iterations 0-4, type B for iterations 5-6, and type C from iteration 7 onwards.
        """
        if len(model_types) != len(num_iterations_per_model) + 1:
            raise Exception(
                f"The number of model types provided ({len(model_types)}) does not match the provided list of "
                f"{len(num_iterations_per_model)} model switch points. For each model type, except the last one, "
                f"the number of iterations for this model type to be used must be specified.")
        self.model_types = model_types
        self.num_iterations_per_model = num_iterations_per_model
        self.switch_points = np.cumsum(num_iterations_per_model)

    def get_model_type(self, iteration_num: int) -> ModelType:
        for model_type, switch_point in zip(self.model_types, self.switch_points):
            if iteration_num < switch_point:
                return model_type
        return self.model_types[-1]

    def get_all_model_types(self):
        return set(self.model_types)

    def get_name(self):
        name = ""
        for model_type, n_iter in zip(self.model_types, self.num_iterations_per_model):
            name += f'{model_type.name}x{n_iter}-'
        name += f"{self.model_types[-1].name}"
        return name



#
#  Copyright (c) 2022 IBM Corp.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import random

from typing import Sequence

from label_sleuth.active_learning.core.active_learning_api import ActiveLearner
from label_sleuth.data_access.core.data_structs import TextElement
from label_sleuth.models.core.prediction import Prediction


class RandomSampling(ActiveLearner):
    """
    This module assigns random (but reproducible) scores for the given examples. Thus, using this active learning module
    equates to randomly selecting the next examples for labeling.
    """
    def get_per_element_score(self, candidate_text_elements: Sequence[TextElement],
                              candidate_text_element_predictions: Sequence[Prediction], workspace_id: str,
                              dataset_name: str, category_id: int) -> Sequence[float]:
        # The scores/selected elements are reproducible given the same *candidate_text_elements* -
        # i.e., if the same set of elements in the dataset was labeled, the list *candidate_text_elements*
        # of unlabeled elements would be identical and the same scores/elements will be returned
        rand = random.Random(0)
        return [rand.random() for _ in range(len(candidate_text_element_predictions))]



import ast
import dataclasses
import logging
import os
import time

from typing import Dict, Tuple, Type

import ujson

from label_sleuth.models.core.prediction import Prediction


def load_model_prediction_store_from_disk(path_to_store: str, prediction_class: Type[Prediction]) \
        -> Dict[Tuple, Prediction]:
    """
    Load predictions for a particular model from a json file
    :param path_to_store: path to a predictions json
    :param prediction_class: class of the model prediction object. Usually of type Prediction.
    :return: a mapping from store keys in the form (('text', element_text),) to *prediction_class* objects.
    """

    if not os.path.isfile(path_to_store):
        return {}
    with open(path_to_store) as reader:
        model_prediction_store = ujson.load(reader)
    res =  {k: prediction_class(**v) for k, v in model_prediction_store.items()}

    return res

def save_model_prediction_store_to_disk(path_to_store: str, model_cache: Dict[Tuple, Prediction]):
    """
    Convert a dict of model predictions to json and save to disk
    :param path_to_store: path to a predictions json
    :param model_cache: a mapping from store keys in the form (('text', element_text),) to Prediction dataclass objects
    (or classes inheriting from the Prediction dataclass).
    """
    os.makedirs(os.path.dirname(path_to_store), exist_ok=True)
    start = time.time()
    model_cache = {k: dataclasses.asdict(v) for k, v in model_cache.items()}
    with open(path_to_store, "w") as output_file:
        output_file.write(ujson.dumps(model_cache))
    end = time.time()
    logging.info(f"saving {len(model_cache)} items to model prediction store on disk took {end-start}")

